from collections import OrderedDict
from io import BytesIO
from typing import NamedTuple, Dict

from .containers import parse_field_anystring, parse_boolean_table_field, STRING_ARGS, FIELD_ARGS, parse_field_value, \
    FieldValue, build_boolean_table_field
from .strings import parse_shortstr, build_shortstr, build_longstr
from .structs import parse_long_uint, pack_long_uint


PEER_PROPERTIES_FIELDS = [
    ('host', (parse_field_anystring, FIELD_ARGS)),
    ('product', (parse_field_anystring, FIELD_ARGS)),
    ('version', (parse_field_anystring, FIELD_ARGS)),
    ('platform', (parse_field_anystring, FIELD_ARGS)),
    ('copyright', (parse_field_anystring, FIELD_ARGS)),
    ('information', (parse_field_anystring, FIELD_ARGS)),
    ('capabilities', (parse_boolean_table_field, STRING_ARGS))
]
PEER_PROPERTIES_MAPPING = dict(PEER_PROPERTIES_FIELDS)
PeerProperties = NamedTuple('PeerProperties', [
    ('host', str),
    ('product', str),
    ('version', str),
    ('platform', str),
    ('copyright', str),
    ('information', str),
    ('capabilities', Dict[str, bool]),
    ('others', Dict[str, FieldValue])
])


def parse_peer_properties(bio: BytesIO, encoding: str, rabbitmq: bool) -> PeerProperties:
    length = parse_long_uint(bio)
    others = {}
    table = OrderedDict()
    for key, _ in PEER_PROPERTIES_FIELDS[:-1]:
        table[key] = None
    table['capabilities'] = {}
    end_offset = bio.tell() + length
    while bio.tell() < end_offset:
        name = parse_shortstr(bio, encoding)
        if name in PEER_PROPERTIES_MAPPING:
            parse_field, args_type = PEER_PROPERTIES_MAPPING[name]
            if args_type == FIELD_ARGS:
                value = parse_field(bio, encoding, rabbitmq)
            else:  # STRING_ARGS
                value = parse_field(bio, encoding)
            table[name] = value
        else:
            value = parse_field_value(bio, encoding, rabbitmq)
            others[name] = value
    return PeerProperties(*table.values(), others)


STRING_PEER_PROPERTIES = [
    'host',
    'product',
    'version',
    'platform',
    'copyright',
    'information',
]


def build_peer_properties(bio: BytesIO, peer_properties: PeerProperties, encoding: str) -> None:
    table_bio = BytesIO()
    for name in STRING_PEER_PROPERTIES:
        value = getattr(peer_properties, name)
        if value is not None:
            build_shortstr(table_bio, name, encoding)
            table_bio.write(b'S')  # always use longstring, both RabbitMQ and vanilla AMQP accepts this
            build_longstr(table_bio, value, encoding)
    build_shortstr(table_bio, 'capabilities', encoding)
    build_boolean_table_field(table_bio, peer_properties.capabilities, encoding)
    # XXX: peer_properties.others is currently not used for client properties, but implement it here if it would

    bio.write(pack_long_uint(table_bio.tell()))
    bio.write(table_bio.getvalue())



from io import BytesIO
from struct import pack
from typing import NamedTuple, List

from ammoo.wire import methods as m
from ammoo.wire.low.misc import parse_boolean, pack_boolean
from ammoo.wire.low.peer_properties import parse_peer_properties, PeerProperties, build_peer_properties
from ammoo.wire.low.strings import parse_binary_longstr, parse_shortstr, parse_binary_shortstr, parse_split_longstr, \
    build_shortstr, build_binary_longstr, parse_split_shortstr, build_split_shortstr
from ammoo.wire.low.structs import parse_octet_uint, parse_short_uint, parse_long_uint
from ammoo.wire.typing import VirtualHost, ReplyCode


ConnectionStartParameters = NamedTuple('ConnectionStartParameters', [
    ('version_major', int),
    ('version_minor', int),
    ('server_properties', PeerProperties),
    ('mechanisms', List[str]),
    ('locales', List[str]),
])
ConnectionStartOkParameters = NamedTuple('ConnectionStartOkParameters', [
    ('client_properties', PeerProperties),
    ('mechanism', str),
    ('response', bytes),
    ('locale', str),
])
# Also used for connection.tune-ok
ConnectionTuneParameters = NamedTuple('ConnectionTuneParameters', [
    ('channel_max', int),
    ('frame_max', int),
    ('heartbeat', int),
])
ConnectionOpenParameters = NamedTuple('ConnectionOpenParameters', [
    ('virtualhost', VirtualHost),
    ('capabilities', List[str]),
    ('insist', bool),
])
ConnectionCloseParameters = NamedTuple('ConnectionCloseParameters', [
    ('reply_code', ReplyCode),
    ('reply_text', str),
    ('class_id', int),
    ('method_id', int),
])


def parse_connection_start_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> ConnectionStartParameters:
    return ConnectionStartParameters(
        parse_octet_uint(bio),
        parse_octet_uint(bio),
        parse_peer_properties(bio, encoding, rabbitmq),
        parse_split_longstr(bio, encoding),
        parse_split_longstr(bio, encoding)
    )


def parse_connection_start_ok_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> ConnectionStartOkParameters:
    return ConnectionStartOkParameters(
        parse_peer_properties(bio, encoding, rabbitmq),
        parse_shortstr(bio, encoding),
        parse_binary_longstr(bio),
        parse_shortstr(bio, encoding)
    )


def pack_connection_start_ok_parameters(
        client_properties: PeerProperties, mechanism: str, response: bytes, locale: str, encoding: str) -> bytes:
    bio = BytesIO()
    build_peer_properties(bio, client_properties, encoding)
    build_shortstr(bio, mechanism, encoding),
    build_binary_longstr(bio, response),
    build_shortstr(bio, locale, encoding)
    return bio.getvalue()


# used for connection.secure-ok too
def parse_connection_secure_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> bytes:
    return parse_binary_longstr(bio)  # value is the response parameter


def parse_connection_tune_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> ConnectionTuneParameters:
    return ConnectionTuneParameters(
        parse_short_uint(bio),
        parse_long_uint(bio),  # check somewhere this is at least 4k
        parse_short_uint(bio)
    )


def pack_connection_tune_ok_parameters(channel_max: int, frame_max: int, heartbeat: int) -> bytes:
    return pack('!HIH', channel_max, frame_max, heartbeat)


def parse_connection_open_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> ConnectionOpenParameters:
    return ConnectionOpenParameters(
        parse_shortstr(bio, encoding),
        parse_split_shortstr(bio, encoding),
        parse_boolean(bio)
    )


def pack_connection_open_parameters(
        virtualhost: VirtualHost, capabilities: List[str], insist: bool, encoding: str) -> bytes:
    bio = BytesIO()
    build_shortstr(bio, virtualhost, encoding)
    build_split_shortstr(bio, capabilities, encoding),
    bio.write(pack_boolean(insist))
    return bio.getvalue()


def parse_connection_open_ok_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> None:
    parse_binary_shortstr(bio)  # reserved value, throw away
    return None


def parse_connection_close_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> ConnectionCloseParameters:
    return ConnectionCloseParameters(
        parse_short_uint(bio),
        parse_shortstr(bio, encoding),
        parse_short_uint(bio),
        parse_short_uint(bio)
    )


CONNECTION_METHOD_PARAMETER_PARSERS = {
    m.METHOD_CONNECTION_START: parse_connection_start_parameters,
    m.METHOD_CONNECTION_START_OK: parse_connection_start_ok_parameters,
    m.METHOD_CONNECTION_SECURE: parse_connection_secure_parameters,  # value is the challenge parameter
    m.METHOD_CONNECTION_SECURE_OK: parse_connection_secure_parameters,
    m.METHOD_CONNECTION_TUNE: parse_connection_tune_parameters,
    m.METHOD_CONNECTION_TUNE_OK: parse_connection_tune_parameters,
    m.METHOD_CONNECTION_OPEN: parse_connection_open_parameters,
    m.METHOD_CONNECTION_OPEN_OK: parse_connection_open_ok_parameters,
    m.METHOD_CONNECTION_CLOSE: parse_connection_close_parameters,
    m.METHOD_CONNECTION_CLOSE_OK: None,
}
CLEAN_CONNECTION_CLOSE_PARAMETERS = pack('!HBHH', 0, 0, 0, 0)



FRAME_TYPE_HEARTBEAT = 8


class HeartbeatFrame:
    __slots__ = ()
    frame_type = FRAME_TYPE_HEARTBEAT

    def __init__(self):
        pass

    def __str__(self):
        return 'Heartbeat frame'


HEARTBEAT_FRAME = HeartbeatFrame()  # singleton value



import asyncio
import socket
import ssl as ssl_module
from functools import partial
from typing import Optional
from urllib.parse import urlparse, parse_qs

from ammoo.auth.abc import Authentication
from ammoo.wire.typing import VirtualHost

from ammoo.auth.password import PasswordAuthentication
from ammoo.connection import Connection


DEFAULT_FRAME_MAX = 128 * 1024
DEFAULT_HEARTBEAT = None
DEFAULT_LOGIN = 'guest'
DEFAULT_PASSWORD = 'guest'


async def connect(
        url: Optional[str]=None, *,
        host: Optional[str]=None,
        port: Optional[int]=None,
        virtualhost: Optional[VirtualHost]= '/',
        ssl: bool=False,
        ssl_context: Optional[ssl_module.SSLContext]=None,
        heartbeat_interval: Optional[int]=None,
        auth: Optional[Authentication]=None,
        login: Optional[str]=None,
        password: Optional[str]=None,
        frame_max: Optional[int]=None,
        loop: Optional[asyncio.AbstractEventLoop]=None,
        connection_factory=Connection
) -> Connection:
    """
    :param url: URL to connect to. Has the format "amqp[s]://login:password@host:port/virtualhost?parameters". All of \
    the URL components are optional. Any other arguments used override the values in the URL.
    :param host: Server address or hostname. Defaults to "localhost".
    :param port: Server port. Defaults to 5672 for AMQP and 5671 or AMQPS.
    :param virtualhost: Virtual host to open. Defaults to "/".
    :param ssl: Use AMQPS instead of unencrypted AMQP.
    :param ssl_context:
    :param heartbeat_interval: Heartbeat interval in seconds to send to and receive heartbeats from server. By \
    default the value suggested by the server is used. 0 turns heartbeats off.
    :param auth: Authentication/AuthenticationMechanism instance to use instead of default PasswordAuthentication with
    the login and password arguments.
    :param login: Username to use for plain password authentication. Defaults to "guest".
    :param password: Password to go along with login. Defaults to "guest".
    :param frame_max: Maximum frame size. Default is 128k, or lower if the server demands it.
    :param loop: Optional asyncio event loop to use
    :param connection_factory: Optional connection class returning function to use instead of Connection
    :return: Connection instance

    """
    if url is not None:
        parsed_url = urlparse(url)
        if not parsed_url.scheme or parsed_url.scheme == 'amqp':
            if ssl is None:
                ssl = False
        elif parsed_url.scheme == 'amqps':
            if ssl is None:
                ssl = True
        else:
            raise ValueError('URL schema must be amqp, amqps or empty')
        if parsed_url.query:
            parsed_query = {key: value[0] for key, value in parse_qs(parsed_url.query, strict_parsing=True).items()}
        else:
            parsed_query = {}
    else:
        parsed_url = None
        parsed_query = None

    if host is None:
        if parsed_url is None or parsed_url.hostname is None:
            host = 'localhost'
        else:
            host = parsed_url.hostname

    if frame_max is None:
        if parsed_query is None or 'frame_max' not in parsed_query:
            frame_max = DEFAULT_FRAME_MAX
        else:
            frame_max = int(parsed_query['frame_max'])

    if ssl:
        if ssl_context is None:
            ssl_context = ssl_module.create_default_context()
    elif ssl_context is not None:
        ssl = True

    if heartbeat_interval is None:
        if parsed_query is None or 'heartbeat_interval' not in parsed_query:
            heartbeat_interval = DEFAULT_HEARTBEAT
        else:
            heartbeat_interval = int(parsed_query['heartbeat_interval'])

    if port is None:
        if parsed_url is None or parsed_url.port is None:
            port = 5671 if ssl else 5672
        else:
            port = parsed_url.port

    if auth is None:
        if login is None:
            if parsed_url is None or parsed_url.username is None:
                login = DEFAULT_LOGIN
            else:
                login = parsed_url.username
        if password is None:
            if parsed_url is None or parsed_url.password is None:
                password = DEFAULT_PASSWORD
            else:
                password = parsed_url.password
        auth = PasswordAuthentication(login, password)

    if loop is None:
        loop = asyncio.get_event_loop()

    protocol_factory = partial(
        connection_factory, virtualhost=virtualhost, loop=loop, heartbeat_interval=heartbeat_interval, auth=auth,
        frame_max=frame_max
    )

    transport, protocol = await loop.create_connection(protocol_factory, host, port, ssl=ssl_context)

    # Without TCP_NODELAY the RPC benchmark gets ~ 80x slower.
    # Not setting TCP_CORK, as that would have a 200ms delay and would probably just slow down the many small packets
    # we send (basically anything except basic.publish + its header and body frames, which we can handle by replacing
    # the 2+#body_frames StreamWriter.write calls with fewer ones if needed)
    transport_socket = transport.get_extra_info('socket')
    if (transport_socket is not None and
            transport_socket.family in {socket.AF_INET, socket.AF_INET6} and
            transport_socket.type == socket.SOCK_STREAM and
            transport_socket.proto == socket.IPPROTO_TCP):
        transport_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
    return protocol



import asyncio
import logging
from collections import deque
from traceback import format_exc
import collections
from typing import Callable, Optional, Union
import sys
from typing import Dict

from ammoo.closeable_queue import CloseableQueue, QueueClosed
from ammoo.exceptions.channel import ChannelClosed

from ammoo.exceptions.connection import UnexpectedFrame, ConnectionClosed
from ammoo.exceptions.consumer import ConsumerCancelled
from ammoo.message import Message, DeliverMessage
from ammoo.wire.class_and_method import BASIC_CONSUME_CAM, BASIC_CONSUME_OK_CAM, BASIC_CANCEL_CAM, BASIC_CANCEL_OK_CAM
from ammoo.exceptions import consumer as consumer_exceptions
from ammoo.wire.frames import MethodFrame
from ammoo.wire.frames.method.basic import pack_basic_cancel_parameters, pack_basic_consume_parameters
from ammoo.wire.low.strings import pack_shortstr
from ammoo.wire.typing import ConsumerTag, QueueName

__all__ = 'ConsumerIterator', 'Consumer'

logger = logging.getLogger(__name__)


class ConsumerIterator:
    def __init__(
            self,
            message_queue: CloseableQueue,
            get_closing_exc: Callable[[], Union[None, ConsumerCancelled, ChannelClosed, ConnectionClosed]]
    ):
        self._message_queue = message_queue
        self._get_closing_exc = get_closing_exc

    if sys.version_info < (3, 5, 2):
        async def __aiter__(self) -> 'ConsumerIterator':
            return self
    else:
        def __aiter__(self) -> 'ConsumerIterator':
            return self

    async def __anext__(self) -> DeliverMessage:
        try:
            return await self._message_queue.get()
        except QueueClosed:
            raise self._get_closing_exc()


collections.AsyncIterator.register(ConsumerIterator)
collections.AsyncIterable.register(ConsumerIterator)


class Consumer:
    def __init__(
            self,
            consumer_tag: ConsumerTag,
            queue_name: QueueName,
            send_method_frame: Callable[[int, int, bytes], None],
            channel_id: int,
            loop: asyncio.AbstractEventLoop,
            on_exit: Callable[[], None],
            no_ack: bool,
            no_local: bool,
            exclusive: bool,
            priority: Optional[int],
            amqp_encoding: str,
            rabbitmq: bool
    ):
        self._consumer_tag = consumer_tag
        self._queue_name = queue_name
        self._send_method_frame = send_method_frame
        self._channel_id = channel_id
        self._entered = False  # set to true in __aenter__
        self._closing_exc = None  # type: Union[None, ConsumerCancelled, ChannelClosed, ConnectionClosed]
        self._on_exit = on_exit  # callback from channel to call when this consumer quits one way or another
        self._message_queue = CloseableQueue(loop=loop)
        self._loop = loop
        self._no_ack = no_ack
        self._no_local = no_local
        self._exclusive = exclusive
        self._priority = priority
        # type: Dict[int, deque[asyncio.Future]]
        self._cam_waiters = {cam: deque() for cam in {BASIC_CONSUME_OK_CAM, BASIC_CANCEL_OK_CAM}}
        self.amqp_encoding = amqp_encoding
        self.rabbitmq = rabbitmq

    @property
    def entered(self) -> bool:
        return self._entered

    @property
    def closing(self) -> bool:
        return self._closing_exc is not None

    def _check_open(self) -> None:
        if not self.entered:
            raise ValueError('Consumer not entered')
        if self.closing:
            raise self._closing_exc

    def release(self, exc: Union[ConsumerCancelled, ChannelClosed, ConnectionClosed]) -> None:
        """Called from channel"""
        if not self.closing:
            self._closing_exc = exc
            self._message_queue.close()
        for futures in self._cam_waiters.values():
            for future in futures:
                if not future.done():
                    future.set_exception(self._closing_exc)
            futures.clear()
        self._on_exit()

    def dispatch_frame(self, frame_or_msg: Union[MethodFrame, DeliverMessage]) -> None:
        if isinstance(frame_or_msg, Message):
            self._message_queue.put_nowait(frame_or_msg)
        else:  # method frame
            cam = frame_or_msg.cam
            if cam == BASIC_CANCEL_CAM:
                self._server_cancel_subscriber(frame_or_msg)
            elif cam in self._cam_waiters:
                try:
                    future = self._cam_waiters[cam].popleft()
                except IndexError as exc:
                    raise UnexpectedFrame(frame_or_msg) from exc
                if not future.done():
                    future.set_result(frame_or_msg)
            else:
                raise UnexpectedFrame(frame_or_msg)

    def _server_cancel_subscriber(self, frame: MethodFrame) -> None:
        """Server is cancelling consumer"""
        if not self.closing:
            if not frame.parameters.no_wait:
                self._send_method_frame(
                    self._channel_id, BASIC_CANCEL_OK_CAM, pack_shortstr(self._consumer_tag, self.amqp_encoding))
            self.release(consumer_exceptions.ServerCancelledConsumer())

    async def __aenter__(self) -> 'Consumer':
        try:
            if self.entered:
                raise ValueError('Consumer already entered')
            if self.closing:
                raise self._closing_exc
            arguments = {}
            if self._priority is not None:
                arguments['x-priority'] = self._priority
            self._send_method_frame(self._channel_id, BASIC_CONSUME_CAM, pack_basic_consume_parameters(
                queue_name=self._queue_name,
                consumer_tag=self._consumer_tag,
                no_local=self._no_local,
                no_ack=self._no_ack,
                exclusive=self._exclusive,
                no_wait=False,
                arguments=arguments,
                encoding=self.amqp_encoding,
                rabbitmq=self.rabbitmq
            ))
            consume_ok_frame = await self._wait_for_cam_frame(BASIC_CONSUME_OK_CAM)
            if consume_ok_frame.parameters != self._consumer_tag:
                raise ValueError('{} != {}'.format(consume_ok_frame.parameters, self._consumer_tag))
            self._entered = True
            return self
        except Exception as exc:
            logger.error('Unhandled exception in Consumer.__aenter__: %s', format_exc())
            self.release(exc=consumer_exceptions.ClientCancelledConsumerError(exc))
            raise

    async def _wait_for_cam_frame(self, cam: int) -> MethodFrame:
        future = self._loop.create_future()
        self._cam_waiters[cam].append(future)
        return await future

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        logger.debug('Consumer.__aexit__: exc_type = %r', exc_type)
        if not self.entered:
            raise ValueError('Consumer not entered')
        try:
            if not self.closing:
                self._check_open()
                self._send_method_frame(self._channel_id, BASIC_CANCEL_CAM, pack_basic_cancel_parameters(
                    consumer_tag=self._consumer_tag,
                    no_wait=False,
                    encoding=self.amqp_encoding
                ))
                await self._wait_for_cam_frame(BASIC_CANCEL_OK_CAM)
                if not self.closing:
                    self._closing_exc = consumer_exceptions.ClientCancelledConsumerOK()
                    self._message_queue.close()
        except Exception as exc:
            if not self.closing:
                self._closing_exc = consumer_exceptions.ClientCancelledConsumerError(exc)
                self._message_queue.close()
            raise
        finally:
            self.release(exc=self._closing_exc)
        return

    if sys.version_info < (3, 5, 2):
        async def __aiter__(self) -> ConsumerIterator:
            self._check_open()
            return ConsumerIterator(self._message_queue, lambda: self._closing_exc)
    else:
        def __aiter__(self) -> ConsumerIterator:
            self._check_open()
            return ConsumerIterator(self._message_queue, lambda: self._closing_exc)


collections.AsyncIterable.register(Consumer)



from asyncio import StreamReader, StreamWriter
from struct import unpack, pack
from typing import Union

from ammoo.wire.constants import FRAME_END
from ammoo.wire.frames.body import BodyFrame, FRAME_TYPE_BODY
from ammoo.wire.frames.header import parse_header_frame, BasicHeaderFrame, FRAME_TYPE_HEADER
from ammoo.wire.frames.heartbeat import HeartbeatFrame, HEARTBEAT_FRAME, FRAME_TYPE_HEARTBEAT
from ammoo.wire.frames.method import parse_method_frame, FRAME_TYPE_METHOD, MethodFrame
from ammoo.wire.low.exceptions import InvalidFrameEnd, InvalidFrameType

Frame = Union[MethodFrame, BasicHeaderFrame, BodyFrame, HeartbeatFrame]


async def parse_frame(reader: StreamReader, encoding: str, rabbitmq: bool) -> Frame:
    frame_type, channel_id, payload_size = unpack('!BHI', await reader.readexactly(7))

    if frame_type == FRAME_TYPE_METHOD:
        frame = await parse_method_frame(reader, payload_size, channel_id, encoding, rabbitmq)
    elif frame_type == FRAME_TYPE_BODY:
        frame = BodyFrame(channel_id, payload_size, await reader.readexactly(payload_size))
    elif frame_type == FRAME_TYPE_HEADER:
        frame = await parse_header_frame(reader, payload_size, channel_id, encoding, rabbitmq)
    elif frame_type == FRAME_TYPE_HEARTBEAT:
        frame = HEARTBEAT_FRAME
    else:
        raise InvalidFrameType(frame_type)

    frame_end = await reader.readexactly(1)
    if frame_end != FRAME_END:
        raise InvalidFrameEnd(frame_end)

    return frame


def build_frame(writer: StreamWriter, frame_type: int, channel_id: int, payload):
    payload_size = len(payload)
    writer.write(pack('!BHI', frame_type, channel_id, payload_size))
    writer.write(payload)
    writer.write(FRAME_END)



from datetime import datetime

from ammoo.wire.constants import FRAME_MIN_SIZE
from ammoo.wire.typing import ExchangeName, DeliveryMode, DeliveryTag, QueueName, RoutingKey


def validate_bool(value: bool, name: str) -> None:
    if not isinstance(value, bool):
        raise TypeError('{} is not a bool: {!r}'.format(name, value))


def validate_str(value: str, name: str) -> None:
    if not isinstance(value, str):
        raise TypeError('{} is not a str: {!r}'.format(name, value))


def validate_int(value: int, name: str) -> None:
    if not isinstance(value, int):
        raise TypeError('{} is not an int: {!r}'.format(name, value))


UINT_64_MAX = 2**64 - 1
UINT_32_MAX = 2**32 - 1
UINT_16_MAX = 2**16 - 1
UINT_8_MAX = 2**8 - 1


def validate_short_short_uint(value: int, name) -> None:
    validate_int(value, name)
    if value < 0 or value > UINT_8_MAX:
        raise ValueError('{} not in range for an 8-bit unsigned integer: {}'.format(name, value))


def validate_short_uint(value: int, name: str) -> None:
    validate_int(value, name)
    if value < 0 or value > UINT_16_MAX:
        raise ValueError('{} not in range for a 16-bit unsigned integer: {}'.format(name, value))


def validate_long_uint(value: int, name: str) -> None:
    validate_int(value, name)
    if value < 0 or value > UINT_32_MAX:
        raise ValueError('{} not in range for a 32-bit unsigned integer: {}'.format(name, value))


def validate_long_long_uint(value: int, name: str) -> None:
    validate_int(value, name)
    if value < 0 or value > UINT_64_MAX:
        raise ValueError('{} not in range for a 64-bit unsigned integer: {}'.format(name, value))


def validate_delivery_tag(value: DeliveryTag, name: str='delivery_tag') -> None:
    validate_long_long_uint(value, name)


def validate_priority(value: int, name: str='priority') -> None:
    # Note: AMQP 0.9.1 defines this as 0-9, but RabbitMQ allows 0-255
    validate_short_short_uint(value, name)


def validate_delivery_mode(value: DeliveryMode, name: str='delivery_mode') -> None:
    validate_int(value, name)
    if value != 1 and value != 2:
        raise ValueError('{} is not a valid delivery mode (1 or 2)')


def validate_frame_max(value: int, name='frame_max') -> None:
    validate_long_uint(value, name)
    if value < FRAME_MIN_SIZE:
        raise ValueError('{} must be at least {}'.format(name, FRAME_MIN_SIZE))


def validate_timestamp(value: datetime, name: str='timestamp') -> None:
    if not isinstance(value, datetime):
        raise TypeError('{} is not datetime: {!r}'.format(name, value))


def validate_shortstr(value: str, name: str):
    validate_str(value, name)
    if len(value) > 255:
        raise ValueError('{} is over 255 characters: {}'.format(name, value))


def validate_exchange_name(value: ExchangeName, name: str='exchange_name') -> None:
    validate_shortstr(value, name)


def validate_queue_name(value: QueueName, name: str='queue_name') -> None:
    validate_shortstr(value, name)


def validate_routing_key(value: RoutingKey, name: str='routing_key') -> None:
    validate_shortstr(value, name)



import asyncio
import collections
import itertools
import json as json_module
import logging
import typing
from asyncio import QueueFull
from collections import Mapping, OrderedDict, deque
from datetime import datetime
from functools import partial
from typing import Dict, Tuple, Iterator, Union, Optional

from ammoo._utils import add_int_arg, add_routing_keys_header
from ammoo.consumer import Consumer
from ammoo.exceptions import channel as channel_exceptions
from ammoo.exceptions.channel import ChannelClosed, PublishNack, ChannelInactive
from ammoo.exceptions.connection import UnexpectedFrame, ConnectionClosed
from ammoo.message import PartialMessage, AbstractMessage, GetMessage, ReturnMessage
from ammoo.validatation import validate_bool, validate_delivery_tag, validate_exchange_name, validate_shortstr, \
    validate_long_uint, \
    validate_short_uint, validate_priority, validate_timestamp, validate_delivery_mode, validate_queue_name, \
    validate_routing_key
from ammoo.wire.class_and_method import CHANNEL_OPEN_CAM, CHANNEL_OPEN_OK_CAM, CHANNEL_CLOSE_CAM, \
    CHANNEL_CLOSE_OK_CAM, EXCHANGE_DECLARE_CAM, EXCHANGE_DECLARE_OK_CAM, EXCHANGE_DELETE_OK_CAM, \
    EXCHANGE_DELETE_CAM, QUEUE_DECLARE_CAM, QUEUE_DECLARE_OK_CAM, QUEUE_DELETE_CAM, QUEUE_DELETE_OK_CAM, \
    QUEUE_PURGE_CAM, QUEUE_PURGE_OK_CAM, QUEUE_BIND_CAM, QUEUE_BIND_OK_CAM, QUEUE_UNBIND_CAM, QUEUE_UNBIND_OK_CAM, \
    BASIC_PUBLISH_CAM, BASIC_ACK_CAM, CONFIRM_SELECT_CAM, CONFIRM_SELECT_OK_CAM, BASIC_GET_CAM, BASIC_GET_EMPTY_CAM, \
    BASIC_GET_OK_CAM, BASIC_CONSUME_OK_CAM, BASIC_CANCEL_OK_CAM, BASIC_DELIVER_CAM, BASIC_CANCEL_CAM, \
    EXCHANGE_BIND_CAM, EXCHANGE_BIND_OK_CAM, BASIC_REJECT_CAM, BASIC_QOS_CAM, BASIC_NACK_CAM, BASIC_QOS_OK_CAM, \
    BASIC_RETURN_CAM, BASIC_RECOVER_CAM, BASIC_RECOVER_OK_CAM, CHANNEL_FLOW_CAM, CHANNEL_FLOW_OK_CAM
from ammoo.wire.frames import FRAME_TYPE_BODY, FRAME_TYPE_HEADER, FRAME_TYPE_METHOD, MethodFrame
from ammoo.wire.frames.header import pack_basic_header_frame_payload, BasicHeaderProperties
from ammoo.wire.frames.method.basic import pack_basic_publish_parameters, pack_basic_get_parameters, \
    pack_basic_ack_parameters, pack_basic_nack_parameters, pack_basic_reject_parameters, pack_basic_qos_parameters
from ammoo.wire.frames.method.channel import CLEAN_CHANNEL_CLOSE_PARAMETERS_DATA
from ammoo.wire.frames.method.exchange import pack_exchange_declare_parameters, pack_exchange_delete_parameters, \
    pack_exchange_bind_parameters
from ammoo.wire.frames.method.queue import QueueDeclareOkParameters, pack_queue_declare_parameters, \
    pack_queue_delete_parameters, pack_queue_purge_parameters, pack_queue_bind_parameters, pack_queue_unbind_parameters
from ammoo.wire.low.containers import FieldValue
from ammoo.wire.low.misc import BOOLEAN_FALSE_DATA, pack_boolean
from ammoo.wire.low.strings import EMPTY_SHORTSTR_DATA
from ammoo.wire.typing import MessageCount, ConsumerTag, ExchangeName, ExchangeType, RoutingKey, QueueName, \
    DeliveryMode, DeliveryTag, EncodeJson, OrderedDictT, Deque

__all__ = 'Channel',

logger = logging.getLogger(__name__)


_JSON_COMPACT_SEPARATORS = (',', ':')

class _DefaultArgumentValue:
    """For typing. Using object() singleton default argument values makes any value acceptable."""


_NO_JSON = _DefaultArgumentValue()  # can't use None for default argument value, as that's null for json


def _publish_encoding(
        body: Union[str, bytes, bytearray, None],
        json: Union[EncodeJson, _DefaultArgumentValue],
        content_encoding: Union[None, str, bool],
        content_type: Union[None, str, bool],
        encoding: Optional[str],
        default_encoding: str
) -> Tuple[Union[bytes, bytearray], Optional[str], Optional[str]]:
    """Helper method for Channel.publish(). Encodes body/json, and sets content_encoding and content_type based
    on (in this order) the encoding argument, the content_encoding argument or the channel's default"""
    # validate body/json and set flags whether we need to do encode str -> bytes later
    if body is None:
        if json is _NO_JSON:
            raise TypeError('Either body or json must be given')
        encode_text = True
        json_set = True
    else:
        if isinstance(body, str):
            encode_text = True
        elif isinstance(body, (bytes, bytearray)):
            encode_text = False
        else:
            raise TypeError('body must be an str, bytes or bytearray instance, not: {!r}'.format(body))
        json_set = False

    # handle content_encoding
    if content_encoding is None:
        pass
    elif isinstance(content_encoding, bool):
        if not encode_text:
            raise TypeError('boolean value for content_encoding cannot be used when body is bytes/bytearray')
        if content_encoding:
            content_encoding = default_encoding if encoding is None else encoding
        else:
            content_encoding = None
    else:
        validate_shortstr(content_encoding, 'content_encoding')
        if encoding is None:
            encoding = content_encoding
        elif encoding != content_encoding:
            logger.debug(
                'Connection.publish argument encoding = %r is different from content_encoding = %r',
                encoding, content_encoding
            )
    if encode_text and encoding is None:
        encoding = default_encoding if content_encoding is None else content_encoding

    # handle content_type
    if content_type is not None:
        if isinstance(content_type, bool):
            if content_type:
                if json_set:
                    content_type = 'application/json'
                elif encode_text:
                    content_type = 'text/plain'
                else:
                    content_type = 'application/octet-stream'
            else:
                content_type = None
        else:
            validate_shortstr(content_type, 'content_type')

    if json_set:
        body = json_module.dumps(json, separators=_JSON_COMPACT_SEPARATORS)
    if encode_text:
        body = body.encode(encoding)

    return body, content_encoding, content_type


# Passed to Message, to avoid it having a direct reference to the channel
_ChannelToolbox = collections.namedtuple('_ChannelToolbox', 'server_has_capability ack reject publish default_encoding')


class Channel:
    def __init__(
            self, channel_id, send_frame, send_method_frame, loop, on_exit, frame_max, server_has_capability,
            drain_writer, prefetch_size, prefetch_count, body_encoding, amqp_encoding, rabbitmq):
        self._channel_id = channel_id
        # set to true in __aenter__
        self._entered = False  # type: bool
        self._closing_exc = None  # type: Union[None, ChannelClosed, ConnectionClosed]
        self._send_frame = send_frame
        self._send_method_frame = send_method_frame
        self._loop = loop
        self._on_exit = on_exit # callback from connection to call when this channel quits one way or another
        self._frame_max = frame_max
        self._publish_confirm_mode = False  # type: bool
        self._delivery_tag_iter = None  # type: Optional[Iterator[DeliveryTag]]
        self._consumer_tag_iter = map(lambda n: '{:x}'.format(n), itertools.count(0))  # type: Iterator[ConsumerTag]
        self._server_has_capability = server_has_capability
        self._consumers = {}  # consumer_tag -> Consumer()
        self._drain_writer = drain_writer
        self._prefetch_size = prefetch_size
        self._prefetch_count = prefetch_count
        self._returned_messages_queue = asyncio.Queue(maxsize=20, loop=self._loop)
        self.body_encoding = body_encoding
        self.amqp_encoding = amqp_encoding
        self.rabbitmq = rabbitmq
        # set to False on receiving a channel.flow from server
        self._active = True  # type: bool
        # passed to Message(), to avoid giving them a direct reference to self, or to pass n+1 args every time
        self._toolbox = _ChannelToolbox(
            self._server_has_capability, self.ack, self.reject, self.publish, self.body_encoding)
        self._partial_message = PartialMessage(self._toolbox)

        self._cam_waiters = {
            cam: deque() for cam in {
                CHANNEL_OPEN_OK_CAM, CHANNEL_CLOSE_OK_CAM, EXCHANGE_DECLARE_OK_CAM, EXCHANGE_DELETE_OK_CAM,
                EXCHANGE_BIND_OK_CAM, QUEUE_DECLARE_OK_CAM, QUEUE_DELETE_OK_CAM, QUEUE_PURGE_OK_CAM,
                QUEUE_BIND_OK_CAM, QUEUE_UNBIND_OK_CAM, CONFIRM_SELECT_OK_CAM, BASIC_QOS_OK_CAM, BASIC_RECOVER_OK_CAM
            }
        }  # type: Dict[int, Deque[asyncio.Future]]
        get_ok_or_empty_deque = deque()
        for cam in BASIC_GET_OK_CAM, BASIC_GET_EMPTY_CAM:
            self._cam_waiters[cam] = get_ok_or_empty_deque
        self._pending_acks = OrderedDict()  # type: OrderedDictT[DeliveryTag, asyncio.Future]

    async def get_next_returned_message(self) -> ReturnMessage:
        """Returns the next basic.return message we have"""
        self._check_open()
        message = await self._returned_messages_queue.get()
        if message is None:
            raise self._closing_exc
        return message

    @property
    def entered(self):
        return self._entered

    @property
    def closing(self):
        return self._closing_exc is not None

    @property
    def channel_id(self):
        return self._channel_id

    def dispatch_frame(self, frame):
        assert frame.channel_id == self._channel_id
        frame_type = frame.frame_type
        if frame_type == FRAME_TYPE_METHOD:
            cam = frame.cam
            if cam in (BASIC_GET_OK_CAM, BASIC_DELIVER_CAM, BASIC_RETURN_CAM):
                self._partial_message.feed_method_frame(frame)
                # XXX: for all other method frames than these, clear the current partial message
            elif cam == BASIC_ACK_CAM or cam == BASIC_NACK_CAM:
                if not self._publish_confirm_mode:
                    logger.warning('Broker sent basic.ack, but we\'re not in publish confirm mode')
                    raise UnexpectedFrame(frame)
                parameters = frame.parameters
                ack_delivery_tag = parameters.delivery_tag
                multiple = parameters.multiple
                # basic.nack has parameter requeue too, but spec tells to ignore it
                if multiple:
                    if cam == BASIC_ACK_CAM:
                        def finish_future(future: asyncio.Future) -> None:
                            if not future.done():
                                future.set_result(None)
                    else:
                        nack_exc = PublishNack()
                        def finish_future(future):
                            if not future.done():
                                future.set_exception(nack_exc)
                    logger.debug('Got multiple ack for delivery tag %s', ack_delivery_tag)
                    while True:
                        try:
                            pending_delivery_tag, future = self._pending_acks.popitem(False)  # False ~ popleft
                        except KeyError as exc:
                            logger.warning(
                                'Broker sent basic.ack for delivery tag %s, but none are pending', ack_delivery_tag)
                            raise UnexpectedFrame(frame) from exc
                        if ack_delivery_tag < pending_delivery_tag:
                            logger.warning(
                                'Broker sent basic.ack for delivery-tag %s with multiple, but that tag if not pending',
                                ack_delivery_tag
                            )
                            raise UnexpectedFrame(frame)
                        logger.debug('Multiple ack -> resolving %s', pending_delivery_tag)
                        finish_future(future)
                        if ack_delivery_tag == pending_delivery_tag:
                            break
                else:
                    logger.debug('Got single ack for delivery tag %s', ack_delivery_tag)
                    try:
                        future = self._pending_acks[ack_delivery_tag]
                    except KeyError as exc:
                        raise UnexpectedFrame(frame) from exc
                    if not future.done():
                        if cam == BASIC_ACK_CAM:
                            future.set_result(None)
                        else:
                            future.set_exception(PublishNack())
            elif cam == CHANNEL_CLOSE_CAM:
                self._send_method_frame(self._channel_id, CHANNEL_CLOSE_OK_CAM, b'')
                if not self.closing:
                    self.release(channel_exceptions.ServerClosedChannel(frame.parameters))
            elif cam == CHANNEL_FLOW_CAM:
                active = frame.parameters
                self._active = active
                logger.debug('Channel active state changed to %s', active)
                self._send_method_frame(self._channel_id, CHANNEL_FLOW_OK_CAM, pack_boolean(active))
            elif cam == BASIC_CANCEL_CAM:
                consumer_tag = frame.parameters.consumer_tag
                self._dispatch_to_consumer(frame, consumer_tag)
            elif cam == BASIC_CANCEL_OK_CAM or cam == BASIC_CONSUME_OK_CAM:
                consumer_tag = frame.parameters
                self._dispatch_to_consumer(frame, consumer_tag)
            else:
                if cam in self._cam_waiters:
                    try:
                        future = self._cam_waiters[cam].popleft()
                    except IndexError as exc:
                        raise UnexpectedFrame(frame) from exc
                    if not future.done():
                        future.set_result(frame)
                else:
                    raise UnexpectedFrame(frame)
        elif frame_type == FRAME_TYPE_HEADER:
            message = self._partial_message.feed_header_frame(frame)
            if message is not None:
                self._dispatch_message(message)
        elif frame_type == FRAME_TYPE_BODY:
            message = self._partial_message.feed_body_frame(frame)
            if message is not None:
                self._dispatch_message(message)
        else:
            raise UnexpectedFrame(frame)

    def _dispatch_message(self, message: AbstractMessage):
        cam = message.cam
        if cam == BASIC_GET_OK_CAM:
            try:
                future = self._cam_waiters[cam].popleft()
            except IndexError:
                raise UnexpectedFrame(message)  # XXX: exception expects a frame, not a message?
            if not future.done():
                future.set_result(message)
        elif cam == BASIC_DELIVER_CAM:
            consumer_tag = message.consumer_tag
            logger.debug('Dispatching basic.deliver message to consumer with tag %s', consumer_tag)
            self._dispatch_to_consumer(message, consumer_tag)
        elif message.cam == BASIC_RETURN_CAM:
            if self._returned_messages_queue is not None:
                try:
                    self._returned_messages_queue.put_nowait(message)
                except QueueFull:
                    logger.warning('Received a basic.return message, but could not put it to full queue')
        else:
            raise ValueError(message)

    def release(self, exc: Union[ChannelClosed, ConnectionClosed]) -> None:
        """Called from __aenter__/__aexit__ + Connection._close"""
        if self._closing_exc is None:
            self._closing_exc = exc
        for futures in self._cam_waiters.values():
            for future in futures:
                if not future.done():
                    future.set_exception(self._closing_exc)
            futures.clear()
        for future in self._pending_acks.values():
            if not future.done():
                future.set_exception(self._closing_exc)
        self._pending_acks.clear()
        for consumer in list(self._consumers.values()):
            consumer.release(self._closing_exc)
        self._returned_messages_queue.put_nowait(None)  # signals we're done here
        assert not self._consumers
        self._on_exit()

    def _dispatch_to_consumer(
            self, frame_or_message: Union[AbstractMessage, MethodFrame], consumer_tag: ConsumerTag) -> None:
        try:
            consumer = self._consumers[consumer_tag]
        except KeyError:
            logger.info('Got frame/message for consumer tag %s, but it is not open', consumer_tag)
            raise UnexpectedFrame(frame_or_message)  # XXX: connection's dispatcher expects a frame, not a message
        else:
            consumer.dispatch_frame(frame_or_message)

    async def __aenter__(self):
        try:
            if self.entered:
                raise ValueError('Channel already entered')
            if self.closing:
                raise self._closing_exc
            self._send_method_frame(self._channel_id, CHANNEL_OPEN_CAM, EMPTY_SHORTSTR_DATA)
            await self._wait_for_cam_frame(CHANNEL_OPEN_OK_CAM)
            self._entered = True
            if self._prefetch_size is not None or self._prefetch_count is not None:
                prefetch_size = 0 if self._prefetch_size is None else self._prefetch_size
                prefetch_count = 0 if self._prefetch_count is None else self._prefetch_count
                # global = false, because when qos'ing on RabbitMQ before any consumers exist on channel, the effect
                # is the same as for global = true, and this way the result is the same as on vanilla AMQP
                # (on which global = true means the qos applies on all channels)
                await self.qos(prefetch_size, prefetch_count, global_=False)
            return self
        except Exception as exc:
            self.release(exc=channel_exceptions.ClientClosedChannelError(exc))
            raise

    async def _wait_for_cam_frame(self, cam: int) -> Union[MethodFrame, AbstractMessage]:
        future = self._loop.create_future()
        self._cam_waiters[cam].append(future)
        return await future

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if not self.entered:
            raise ValueError('Channel not entered')
        try:
            if not self.closing:
                self._send_method_frame(self._channel_id, CHANNEL_CLOSE_CAM, CLEAN_CHANNEL_CLOSE_PARAMETERS_DATA)
                self._closing_exc = channel_exceptions.ClientClosedChannelOK()
                await self._wait_for_cam_frame(CHANNEL_CLOSE_OK_CAM)
        finally:
            self.release(exc=self._closing_exc)

    def _check_open(self) -> None:
        """Utility method to check channel is __aentered__ and not closing/-ed. Must be called before sending frames
        to broker, or waiting for receiving them (with eg self._wait_for_cam_frame)"""
        if not self.entered:
            raise ValueError('Channel not entered')
        if self.closing:
            raise self._closing_exc

    async def assert_exchange_exists(self, exchange_name: ExchangeName) -> None:
        """
        Asserts *exchange_name* exists*. Channel will be closed if it does not!

        :param exchange_name: Exchange name
        """
        validate_exchange_name(exchange_name)
        self._check_open()
        self._send_method_frame(
            self._channel_id, EXCHANGE_DECLARE_CAM,
            pack_exchange_declare_parameters(
                exchange_name=exchange_name,
                type_='',
                passive=True,
                durable=False,
                auto_delete=False,
                internal=False,
                no_wait=False,
                arguments={},
                encoding=self.amqp_encoding,
                rabbitmq=self.rabbitmq
            )
        )
        await self._wait_for_cam_frame(EXCHANGE_DECLARE_OK_CAM)  # no parameters returned
        return None

    async def declare_exchange(
            self,
            exchange_name: ExchangeName,
            exchange_type: ExchangeType, *,
            durable: bool=False,
            auto_delete: bool=False,
            alternate_exchange_name: Optional[ExchangeName]=None
    ) -> None:
        """
        Declare exchange.

        :param str exchange_name: Exchange name
        :param str exchange_type: Exchange type: direct, fanout, topic, or headers.
        :param bool durable: Optional: If True, exchange will survive server restart.
        :param bool auto_delete: Optional: If True, the exchange is deleted (after a delay) when the last queue is unbound from it.
        :param alternate_exchange_name: Optional: Alternate exchange name. Messages that can't be routed to any queue are instead published on the alternate exchange (Sets alternate-exchange on the exchange).
        """
        validate_exchange_name(exchange_name)
        validate_bool(durable, 'durable')
        validate_bool(auto_delete, 'auto_delete')
        arguments = {}
        if alternate_exchange_name is not None:
            validate_exchange_name(alternate_exchange_name, 'alternate_exchange_name')
            arguments['alternate-exchange'] = alternate_exchange_name
        self._check_open()
        self._send_method_frame(self._channel_id, EXCHANGE_DECLARE_CAM, pack_exchange_declare_parameters(
            exchange_name=exchange_name,
            type_=exchange_type,
            passive=False,
            durable=durable,
            auto_delete=auto_delete,
            internal=False,
            no_wait=False,
            arguments=arguments,
            encoding=self.amqp_encoding,
            rabbitmq=self.rabbitmq
        ))
        await self._wait_for_cam_frame(EXCHANGE_DECLARE_OK_CAM)  # no parameters returned
        return

    async def delete_exchange(self, exchange_name: ExchangeName, *, if_unused: bool=False) -> None:
        """
        Delete exchange *exchange_name*.

        :param str exchange_name: Exchange name
        :param bool if_unused: Optional: Only delete exchange if it is unused.
        """
        validate_exchange_name(exchange_name)
        validate_bool(if_unused, 'if_unused')
        self._check_open()
        self._send_method_frame(self._channel_id, EXCHANGE_DELETE_CAM, pack_exchange_delete_parameters(
            exchange_name=exchange_name,
            if_unused=if_unused,
            no_wait=False,
            encoding=self.amqp_encoding
        ))
        await self._wait_for_cam_frame(EXCHANGE_DELETE_OK_CAM)  # no parameters returned
        return

    async def bind_exchange(self, destination: ExchangeName, source: ExchangeName, routing_key: RoutingKey) -> None:
        """
        Bind *source* exchange to *destination* exchange for *routing_key*.

        :param str destination: Destination exchange name
        :param str source: Source exchange name
        :param str routing_key: Routing key

        .. note:: RabbitMQ extension. Not supported by standard AMQP.
        """
        validate_exchange_name(destination, 'destination')
        validate_exchange_name(source, 'source')
        validate_routing_key(routing_key)
        if not self._server_has_capability('exchange_exchange_bindings'):
            raise ValueError('Server does not support capability exchange_exchange_bindings')
        self._check_open()
        self._send_method_frame(self._channel_id, EXCHANGE_BIND_CAM, pack_exchange_bind_parameters(
            destination=destination,
            source=source,
            routing_key=routing_key,
            no_wait=False,
            arguments={},
            encoding=self.amqp_encoding,
            rabbitmq=self.rabbitmq
        ))
        await self._wait_for_cam_frame(EXCHANGE_BIND_OK_CAM)  # no parameters returned
        return

    async def assert_queue_exists(self, queue_name: QueueName) -> QueueDeclareOkParameters:
        """
        Asserts *queue_name* exists*. Channel is closed by the server if it does not!

        :param queue_name: Queue name
        :rtype: ~ammoo.wire.frames.method.queue.QueueDeclareOkParameters
        """
        """If queue exists, return its parameters. If not, the channel is closed and ServerClosedChannel raised"""
        validate_queue_name(queue_name)
        self._check_open()
        self._send_method_frame(self._channel_id, QUEUE_DECLARE_CAM, pack_queue_declare_parameters(
            queue_name=queue_name,
            passive=True,
            durable=False,
            exclusive=False,
            auto_delete=False,
            no_wait=False,
            arguments={},
            encoding=self.amqp_encoding,
            rabbitmq=self.rabbitmq
        ))
        frame = await self._wait_for_cam_frame(QUEUE_DECLARE_OK_CAM)
        return frame.parameters

    async def declare_queue(
            self,
            queue_name: QueueName, *,
            exclusive: bool=False, durable: bool=False, auto_delete: bool=False,
            ttl: Optional[int]=None, expires: Optional[int]=None,
            max_length: Optional[int]=None, max_length_bytes: Optional[int]=None,
            dead_letter_exchange: Optional[ExchangeName]=None, dead_letter_routing_key: Optional[RoutingKey]=None,
            max_priority: Optional[int]=None
    ) -> QueueDeclareOkParameters:
        """
        Declare a queue named *queue_name*.

        :param str queue_name: Queue name
        :param bool exclusive: Optional: If True, the queue can be accessed only by this connection and is deleted \
        when connection is closed.
        :param bool durable: Optional: If True, the queue will be marked as durable, surviving server restarts.
        :param bool auto_delete: Optional: If True, the queue is deleted when all consumers are finished using it.
        :param int ttl: Optional: Messages die after this number of milliseconds, if no one has consumed them first \
        (Sets x-message-ttl on the queue).
        :param int expires: Optional: Milliseconds queue is unused before it is deleted (Sets x-expires on the queue).
        :param int max_length: Optional: Maximum number of messages in the queue before the oldest will die (Sets \
        x-max-length on the queue).
        :param int max_length_bytes: Optional: Maximum number of message bytes in the queue before the oldest will \
        die (Sets x-max-length on the queue).
        :param str dead_letter_exchange: Optional: Name of dead letter exchange. The queue's dead messages are routed \
        here (Sets x-dead-letter-exchange on the queue).
        :param dead_letter_routing_key: Optional: Override dead messages' routing key when routing them to \
        *dead_letter_exchange* (Sets x-dead-letter-routing-key on the queue).
        :param int max_priority: Optional: Queue's maximum priority (Sets x-max-priority).
        :rtype: ~ammoo.wire.frames.method.queue.QueueDeclareOkParameters
        """
        validate_queue_name(queue_name)
        validate_bool(exclusive, 'exclusive')
        validate_bool(durable, 'durable')
        validate_bool(auto_delete, 'auto_delete')
        arguments = {}
        if ttl is not None:
            add_int_arg('ttl', ttl, 'x-message-ttl', arguments, True)
        if expires is not None:
            add_int_arg('expires', expires, 'x-expires', arguments, False)
        if max_length is not None:
            add_int_arg('max_length', max_length, 'x-max-length', arguments, True)
        if max_length_bytes is not None:
            add_int_arg('max_length_bytes', max_length_bytes, 'x-max-length-bytes', arguments, True)
        if dead_letter_exchange is not None:
            validate_exchange_name(dead_letter_exchange, 'dead_letter_exchange')
            arguments['x-dead-letter-exchange'] = dead_letter_exchange
        if dead_letter_routing_key is not None:
            if dead_letter_exchange is None:
                raise TypeError('dead_letter_routing_key can\'t be used without dead_letter_exchange')
            validate_routing_key(dead_letter_routing_key, 'dead_letter_routing_key')
            arguments['x-dead-letter-routing-key'] = dead_letter_routing_key
        if max_priority is not None:
            validate_priority(max_priority, 'max_priority')
            arguments['x-max-priority'] = max_priority
        self._check_open()
        self._send_method_frame(self._channel_id, QUEUE_DECLARE_CAM, pack_queue_declare_parameters(
            queue_name=queue_name,
            passive=False,
            durable=durable,
            exclusive=exclusive,
            auto_delete=auto_delete,
            no_wait=False,
            arguments=arguments,
            encoding=self.amqp_encoding,
            rabbitmq=self.rabbitmq
        ))
        frame = await self._wait_for_cam_frame(QUEUE_DECLARE_OK_CAM)
        return frame.parameters

    async def delete_queue(self, queue_name: QueueName, *, if_unused: bool=False, if_empty: bool=False) -> MessageCount:
        """
        Delete a queue named *queue_name*. If the queue does not exist, the method merely asserts it is not there.

        :param queue_name: Queue name
        :param bool if_unused: Optional: Only delete queue if it has no consumers.
        :param bool if_empty: Optional: Only delete queue if it has no messages.
        :return: Number of messages in queue before it was deleted
        """
        validate_queue_name(queue_name)
        validate_bool(if_unused, 'if_unused')
        validate_bool(if_empty, 'if_empty')
        self._check_open()
        self._send_method_frame(self._channel_id, QUEUE_DELETE_CAM, pack_queue_delete_parameters(
            queue_name=queue_name,
            if_unused=if_unused,
            if_empty=if_empty,
            no_wait=False,
            encoding=self.amqp_encoding
        ))
        frame = await self._wait_for_cam_frame(QUEUE_DELETE_OK_CAM)
        return frame.parameters

    async def purge_queue(self, queue_name: QueueName) -> MessageCount:
        """
        Purges a queue of messages, emptying it.

        :param str queue_name: Queue name
        :return: Number of messages in queue before it was purged
        """
        validate_queue_name(queue_name)
        self._check_open()
        self._send_method_frame(self._channel_id, QUEUE_PURGE_CAM, pack_queue_purge_parameters(
            queue_name=queue_name,
            no_wait=False,
            encoding=self.amqp_encoding
        ))
        frame = await self._wait_for_cam_frame(QUEUE_PURGE_OK_CAM)
        return frame.parameters

    async def bind_queue(self, queue_name: QueueName, exchange_name: ExchangeName, routing_key: RoutingKey):
        """
        Bind *queue_name* to *exchange_name* for *routing_key*.

        :param str queue_name: Queue name
        :param str exchange_name: Exchange name
        :param routing_key: A :class:`str` is used as a literal routing key, and a :class:`~collections.abc.Mapping` for the headers \
        exchange type.
        """
        validate_queue_name(queue_name)
        validate_exchange_name(exchange_name)
        if isinstance(routing_key, str):
            validate_routing_key(routing_key)
            arguments = {}
        elif isinstance(routing_key, Mapping):
            # NB: https://www.rabbitmq.com/amqp-0-9-1-errata.html 3. Field types. So much for standards.
            arguments = {}
            for argument_name, value in routing_key.items():
                arguments[argument_name] = value
            routing_key = ''
        else:
            raise TypeError('routing_key is not str or Mapping: {!r}'.format(routing_key))
        self._check_open()
        self._send_method_frame(self._channel_id, QUEUE_BIND_CAM, pack_queue_bind_parameters(
            queue_name=queue_name,
            exchange_name=exchange_name,
            routing_key=routing_key,
            no_wait=False,
            arguments=arguments,
            encoding=self.amqp_encoding,
            rabbitmq=self.rabbitmq
        ))
        await self._wait_for_cam_frame(QUEUE_BIND_OK_CAM)  # no parameters returned

    async def unbind_queue(self, queue_name: QueueName, exchange_name: ExchangeName, routing_key: RoutingKey) -> None:
        """
        Unbind *queue_name* from *exchange_name* for *routing_key*. Undoes :meth:`bind_queue`.

        :param str queue_name: Queue name
        :param str exchange_name: Exchange name
        :param routing_key: Same as for :meth:`bind_queue`.
        """
        # TODO: support mapping like bind does
        validate_queue_name(queue_name)
        validate_exchange_name(exchange_name)
        self._check_open()
        self._send_method_frame(self._channel_id, QUEUE_UNBIND_CAM, pack_queue_unbind_parameters(
            queue_name=queue_name,
            exchange_name=exchange_name,
            routing_key=routing_key,
            arguments={},
            encoding=self.amqp_encoding,
            rabbitmq=self.rabbitmq
        ))
        await self._wait_for_cam_frame(QUEUE_UNBIND_OK_CAM)  # no parameters returned
        return

    async def _publish(
            self,
            exchange_name: ExchangeName,
            routing_key: RoutingKey,
            body: Union[bytes, bytearray],
            mandatory: bool,
            immediate: bool,
            properties: Optional[BasicHeaderProperties]
    ) -> None:
        """Utility method for the high level API publish method. Does NOT check the arguments"""
        body_size = len(body)

        # drain writer before sending frames so we don't flood the buffer
        self._check_open()
        await self._drain_writer()
        self._check_open()  # all bets are off after await
        if not self._active:
            # RabbitMQ does not send channel.flow since version 2.0 so this is a bit theoretical. We could also have
            # a mode that awaits until channel is active again
            raise ChannelInactive()

        if self._publish_confirm_mode:
            delivery_tag = next(self._delivery_tag_iter)
        else:
            delivery_tag = None

        # NOTE: if refactored to use writer.drain() (or some other await) between frames, use a mutex so another task
        # on the same channel cannot send frames in between which would be interpreted as cancelling the publishing of
        # this message. Other channels may interleave frames.
        self._send_method_frame(self._channel_id, BASIC_PUBLISH_CAM, pack_basic_publish_parameters(
            exchange_name=exchange_name,
            routing_key=routing_key,
            mandatory=mandatory,
            immediate=immediate,
            encoding=self.amqp_encoding
        ))
        self._send_frame(FRAME_TYPE_HEADER, self.channel_id, pack_basic_header_frame_payload(
            body_size=body_size,
            properties=properties,
            encoding=self.amqp_encoding, rabbitmq=self.rabbitmq
        ))

        offset = 0
        # max payload bytes that can fit into a body frame. 7 is the frame header size and +1 for frame end
        max_payload_size = self._frame_max - 8
        while offset < body_size:
            next_offset = offset + max_payload_size
            if next_offset > body_size:
                next_offset = body_size
            self._send_frame(FRAME_TYPE_BODY, self._channel_id, body[offset:next_offset])
            offset = next_offset

        if self._publish_confirm_mode:
            future = self._loop.create_future()
            self._pending_acks[delivery_tag] = future
            # if server basic.nacks us, this will raise PublishNack
            await future  # no parameters returned
        return

    async def publish(
            self,
            exchange_name: ExchangeName,
            route: Union[RoutingKey, typing.Mapping[str, FieldValue]],
            body: Union[None, str, bytes, bytearray]=None, *,
            json: Union[EncodeJson, _DefaultArgumentValue]=_NO_JSON,
            mandatory: bool=False,
            immediate: bool=False,
            encoding: Optional[str]=None,
            correlation_id: Optional[str]=None,
            reply_to: Optional[str]=None,
            expiration: Union[str, int]=None,
            cc: Optional[RoutingKey]=None,
            bcc: Optional[RoutingKey]=None,
            priority: Optional[int]=None,
            delivery_mode: Optional[DeliveryMode]=None,
            timestamp: Optional[datetime]=None,
            content_encoding: Optional[str]=None,
            content_type: Optional[str]=None,
            message_id: Optional[str]=None,
            type_: Optional[str]=None,
            user_id: Optional[str]=None,
            app_id: Optional[str]=None
    ) -> None:
        """
        Publish a message *body* to *exchange_name* with *route*. *body* and *json* are mutually exclusive, but
        one of them has to be used.

        Publish a binary body with :class:`bytes` or :class:`bytearray`::

            await channel.publish(exchange_name, routing_key, b'binary bytes')
            await channel.publish(exchange_name, routing_key, bytearray(b'binary bytarray'))

        Publish a :class:`str`::

            # encoded to bytes with channel's default encoding
            await channel.publish(exchange_name, routing_key, 'text string')
            # use non-default encoding
            await channel.publish(exchange_name, routing_key, 'text string', encoding='iso-8859-1')

        Serialize JSON into body::

            await channel.publish(exchange_name, routing_key, json={'key': 123})

        Set the content-encoding property::

            # body will also be encoded with iso-8859-1 instead of channel's default encoding
            await channel.publish(exchange_name, routing_key, 'some text', content_encoding='iso-8859-1'})
            # content-encoding property is set to channel's default encoding
            await channel.publish(exchange_name, routing_key, 'some text', content_encoding=True})
            # can be used for json too
            await channel.publish(exchange_name, routing_key, json={'key': 123}, content_encoding='iso-8859-1'})

        Set the content-type property::

            await channel.publish(exchange_name, routing_key, b'binary data', content_type=True})  # bytes body -> content-type is set to application/octet-stream
            await channel.publish(exchange_name, routing_key, 'some text', content_type=True})  # str body -> content-type is set to text/plain
            await channel.publish(exchange_name, routing_key, json={'key': 123}, content_type=True})  # json -> content-type is set to application/json
            await channel.publish(exchange_name, routing_key, body, content_type='application/acme-2000'})  # or a regular str

        :param str exchange_name: Exchange name. Use an empty string for the default exchange.
        :param route: A :class:`str` is used as a literal routing key, while a :class:`~collections.abc.Mapping` is used for the headers exchange type.
        :param body: Message body. :class:`str`, :class:`bytes` or :class:`bytearray`. If the *json* keyword argument is used, body may be omitted.
        :param json: Optional: Object to serialize as JSON into body. Cannot be used at the same time as the body argument.
        :param bool mandatory: Optional: When True, messages the cannot be routed to a queue are returned back to the client.
        :param bool immediate: Optional: When True, messages that are not routed to a consumer immediately are returned back to the client.
        :param str encoding: Optional: Encode :class:`str` body to bytes with this encoding.
        :param str correlation_id: Optional: Correlation-id property.
        :param str reply_to: Optional: Reply-to property.
        :param expiration: Optional: Message expiration property, usually in milliseconds. Messages die if they are \
        not consumed from queue within this TTL. :class:`int` or :class:`str`.
        :param cc: Optional: Additional routing keys to use when routing message to queues.
        :param bcc: Optional: Like cc, but bcc will be removed from message before delivery.
        :param int priority: Optional: priority property.
        :param int delivery_mode: Optional: delivery-mode property 1 for non-persistent or 2 for persistent.
        :param datetime timestamp: Optional: Message timestamp property. If time zone is not set, UTC is assumed.
        :param content_encoding: Optional: content-encoding property. A :class:`str`, or if a :class:`bool` True, the \
        value of the *encoding* argument or the channel's default.
        :param content_type: Optional: content-type property. A :class:`str`, or if a :class:`bool` True, set \
        `application/octet-stream` if body is :class:`bytes` or :class:`bytearray`, `text/plain` if it's :class:`str`, \
        and `application/json` if the *json* argument was used.
        :param str message_id: Optional: message-id property.
        :param str type_: Optional: type property.
        :param str user_id: Optional: user-id property.
        :param str app_id: Optional: app-id property.

        .. note:: cc and bcc are RabbitMQ extensions. Not supported by standard AMQP.
        """
        body, content_encoding, content_type = _publish_encoding(
            body, json, content_encoding, content_type, encoding, self.body_encoding)
        validate_exchange_name(exchange_name)

        headers = {}
        # NB: if route is a str, it's interpreted as a regular routing key. If it's a dict, routing key will be
        # set to the empty string and the dictionary used for the headers property
        if isinstance(route, str):
            validate_routing_key(route)
        elif isinstance(route, Mapping):
            for name, value in route.items():
                headers[name] = value
            route = ''
        else:
            raise TypeError('route is not a str or a Mapping: {!r}'.format(route))
        validate_bool(immediate, 'immediate')
        if expiration is not None:
            if isinstance(expiration, int):
                if expiration < 0:
                    raise ValueError('expiration is negative: {}'.format(expiration))
                expiration = str(expiration)
            elif not isinstance(expiration, str):
                raise TypeError('expiration is not int or str: {!r}'.format(expiration))
            validate_shortstr(expiration, 'expiration')
        if cc is not None:
            add_routing_keys_header(cc, 'cc', headers, 'CC')
        if bcc is not None:
            add_routing_keys_header(bcc, 'bcc', headers, 'BCC')
        if priority is not None:
            validate_priority(priority)
        if delivery_mode is not None:
            validate_delivery_mode(delivery_mode)
        if timestamp is not None:
            validate_timestamp(timestamp)
        if message_id is not None:
            validate_shortstr(message_id, 'message_id')
        if type_ is not None:
            validate_shortstr(type_, 'type_')
        if user_id is not None:
            validate_shortstr(user_id, 'user_id')
        if app_id is not None:
            validate_shortstr(app_id, 'app_id')
        if (mandatory or immediate) and not self._publish_confirm_mode:
            # without publish confirms we don't get acks, and then we don't know if the return was for us.
            raise ValueError(
                'To use mandatory or immediate, channel has to be set into publisher confirm mode with ' +
                'select_confirm() first')
        if not headers:
            headers = None

        properties = BasicHeaderProperties(
            content_type, content_encoding, headers, delivery_mode, priority, correlation_id, reply_to,
            expiration, message_id, timestamp, type_, user_id, app_id, None
        )

        return await self._publish(exchange_name, route, body, mandatory, immediate, properties)

    async def select_confirm(self) -> None:
        """
        Turns publisher confirms on for the channel.

        .. note:: RabbitMQ extension. Not supported by standard AMQP.
        """
        if self._publish_confirm_mode:
            return  # no need to do this twice, though RabbitMQ allows sending many confirm.selects on same channel
        if not self._server_has_capability('publisher_confirms'):
            raise ValueError('Server does not support capability publisher_confirms')
        self._check_open()
        self._send_method_frame(self._channel_id, CONFIRM_SELECT_CAM, BOOLEAN_FALSE_DATA)
        await self._wait_for_cam_frame(CONFIRM_SELECT_OK_CAM)
        self._delivery_tag_iter = itertools.count(1)
        self._publish_confirm_mode = True

    async def get(self, queue_name: QueueName, *, no_ack: bool=False) -> GetMessage:
        """
        Get a message from queue.

        :param str queue_name: Queue name
        :param bool no_ack: Optional: If True, server does not expect message to be acknowledged or rejected.
        :raises EmptyQueue: If there are no messages in queue, EmptyQueue is raised
        :rtype: GetMessage
        """
        validate_queue_name(queue_name)
        validate_bool(no_ack, 'no_ack')
        self._check_open()
        self._send_method_frame(self._channel_id, BASIC_GET_CAM, pack_basic_get_parameters(
            queue_name=queue_name,
            no_ack=no_ack,
            encoding=self.amqp_encoding
        ))
        value = await self._wait_for_cam_frame(BASIC_GET_OK_CAM)  # also resolved by basic.empty
        if isinstance(value, GetMessage):
            # it's basic.get-ok, a message
            return value
        else:
            raise channel_exceptions.EmptyQueue()  # got a basic.empty

    async def ack(self, delivery_tag: DeliveryTag) -> None:
        """
        Acknowledge a message.

        :param int delivery_tag: Delivery tag of message to acknowledge (:attr:`ExpectedMessage.delivery_tag`)

        .. seealso:: :meth:`ExpectedMessage.ack`
        """
        validate_delivery_tag(delivery_tag)
        self._check_open()
        self._send_method_frame(self._channel_id, BASIC_ACK_CAM, pack_basic_ack_parameters(
            delivery_tag=delivery_tag,
            multiple=False
        ))
        return

    async def _nack(self, delivery_tag: DeliveryTag, requeue: bool, multiple: bool) -> None:
        # Unused for now. Does the same as reject when multiple is False.
        validate_delivery_tag(delivery_tag)
        validate_bool(requeue, 'requeue')
        validate_bool(multiple, 'multiple')
        self._check_open()
        self._send_method_frame(self._channel_id, BASIC_NACK_CAM, pack_basic_nack_parameters(
            delivery_tag=delivery_tag,
            multiple=multiple,
            requeue=requeue
        ))
        return

    async def reject(self, delivery_tag: DeliveryTag, requeue: bool) -> None:
        """
        Reject a message. Opposite of :meth:`ack`.

        :param int delivery_tag: Delivery tag of message to reject (:attr:`ExpectedMessage.delivery_tag`)
        :param bool requeue: If True, the server will try to requeue the message. False means the message is discarded or dead-lettered.

        .. seealso:: :meth:`ExpectedMessage.reject`
        """
        validate_delivery_tag(delivery_tag)
        validate_bool(requeue, 'requeue')
        self._check_open()
        self._send_method_frame(self._channel_id, BASIC_REJECT_CAM, pack_basic_reject_parameters(
            delivery_tag=delivery_tag,
            requeue=requeue
        ))
        return

    def _on_consumer_exit(self, consumer_tag: ConsumerTag) -> None:
        """Callback for Consumer to call on its release"""
        if consumer_tag in self._consumers:
            del self._consumers[consumer_tag]

    def consume(
            self,
            queue_name: QueueName, *,
            no_ack: bool=False, no_local: bool=False, exclusive: bool=False,
            priority: Optional[int]=None
    ) -> Consumer:
        """
        Start a new consumer on a queue.

        :param str queue_name: Queue name
        :param bool no_ack: Optional: If True, server does not expect messages delivered to consumer to be \
        acknowledged or rejected.
        :param bool no_local: Optional: If True, the server will not deliver messages to the connection that published them.
        :param bool exclusive: Optional: If True, only this consumer can access the queue.
        :param int priority: Optional: Set consumer priority. Lower priority consumers will receive messages only when higher priority ones are busy (Sets x-priority on the consumer).
        :rtype: Consumer

        .. note:: priority is a RabbitMQ extension
        """
        validate_queue_name(queue_name)
        validate_bool(no_ack, 'no_ack')
        validate_bool(no_local, 'no_local')
        validate_bool(exclusive, 'exclusive')
        if priority is not None:
            validate_priority(priority)
            if not self._server_has_capability('consumer_priorities'):
                raise ValueError('Server does not support capability consumer_priorities')
        self._check_open()
        consumer_tag = next(self._consumer_tag_iter)
        if len(consumer_tag) > 255:
            # Not very likely..
            raise ValueError('Out of consumer tags: {} is too long for shortstr'.format(consumer_tag))

        consumer = Consumer(
            consumer_tag, queue_name, self._send_method_frame, self._channel_id, self._loop,
            partial(self._on_consumer_exit, consumer_tag), no_ack=no_ack,
            no_local=no_local, exclusive=exclusive, priority=priority, amqp_encoding=self.amqp_encoding,
            rabbitmq=self.rabbitmq
        )
        self._consumers[consumer_tag] = consumer
        return consumer

    async def qos(self, prefetch_size: int, prefetch_count: int, global_: bool) -> None:
        """
        Limit how many unacknowledged messages (or message data) will be delivered to consumers. Without `qos`, the
        server will deliver all of the queue's messages to the consumer, possibly causing the consumer to run out of
        memory or starving other consumers of messages.

        :param int prefetch_size: Maximum number of unacknowledged messages server will deliver to a consumer. Zero turns the limit off.
        :param int prefetch_count: Maximum combined size of unacknowledged messages server will deliver to a consumer. Zero turns the limit off.
        :param bool global_: For standard AMQP: True applies the qos to the whole connection, and False to the channel \
        only. For RabbitMQ: True applies the setting to both the channel's current consumers and future ones, while \
        False only applies to the latter.
        """
        validate_long_uint(prefetch_size, 'prefetch_size')
        validate_short_uint(prefetch_count, 'prefetch_count')
        validate_bool(global_, 'global_')
        self._check_open()
        self._send_method_frame(self._channel_id, BASIC_QOS_CAM, pack_basic_qos_parameters(
            prefetch_size=prefetch_size,
            prefetch_count=prefetch_count,
            global_=global_
        ))

        await self._wait_for_cam_frame(BASIC_QOS_OK_CAM)  # no parameters returned
        self._prefetch_size = prefetch_size
        self._prefetch_count = prefetch_count
        return

    async def recover(self, requeue: bool) -> None:
        """
        Ask server to redeliver unacknowledged messages

        :param bool requeue: If False, redeliver messages to the original recipient. If True, the message may be \
        delivered to another recipient.
        """
        validate_bool(requeue, 'requeue')
        self._check_open()
        self._send_method_frame(self._channel_id, BASIC_RECOVER_CAM, pack_boolean(requeue))
        await self._wait_for_cam_frame(BASIC_RECOVER_OK_CAM)  # no parameters returned
        return



from io import BytesIO
from typing import NamedTuple

from ammoo.wire import methods as m
from ammoo.wire.low.containers import parse_field_table, FieldTable, build_field_table
from ammoo.wire.low.misc import parse_booleans, parse_boolean, pack_booleans, pack_boolean
from ammoo.wire.low.strings import parse_binary_shortstr, parse_shortstr, build_shortstr
from ammoo.wire.low.structs import RESERVED_SHORT_UINT_DATA
from ammoo.wire.typing import ExchangeName, ExchangeType, RoutingKey

ExchangeDeclareParameters = NamedTuple('ExchangeDeclareParameters', [
    ('exchange_name', ExchangeName),
    ('type_', ExchangeType),
    ('passive', bool),
    ('durable', bool),
    ('auto_delete', bool),
    ('internal', bool),
    ('no_wait', bool),
    ('arguments', FieldTable),
])
ExchangeDeleteParameters = NamedTuple('ExchangeDeleteParameters', [
    ('exchange_name', ExchangeName),
    ('if_unused', bool),
    ('no_wait', bool),
])
ExchangeBindParameters = NamedTuple('ExchangeBindParameters', [
    ('destination', ExchangeName),
    ('source', ExchangeName),
    ('routing_key', RoutingKey),
    ('no_wait', bool),
    ('arguments', FieldTable),
])


def parse_exchange_declare_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> ExchangeDeclareParameters:
    parse_binary_shortstr(bio)  # reserved
    return ExchangeDeclareParameters(
        parse_shortstr(bio, encoding),
        parse_shortstr(bio, encoding),
        *parse_booleans(bio, 5),
        parse_field_table(bio, encoding, rabbitmq)
    )


def pack_exchange_declare_parameters(
        exchange_name: ExchangeName, type_: ExchangeType, passive: bool, durable: bool, auto_delete: bool,
        internal: bool, no_wait: bool, arguments: FieldTable, encoding: str, rabbitmq: bool) -> bytes:
    bio = BytesIO()
    bio.write(RESERVED_SHORT_UINT_DATA)
    build_shortstr(bio, exchange_name, encoding)
    build_shortstr(bio, type_, encoding)
    bio.write(pack_booleans(passive, durable, auto_delete, internal, no_wait))
    build_field_table(bio, arguments, encoding, rabbitmq)
    return bio.getvalue()


def parse_exchange_delete_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> ExchangeDeleteParameters:
    parse_binary_shortstr(bio)  # reserved
    return ExchangeDeleteParameters(
        parse_shortstr(bio, encoding),
        *parse_booleans(bio, 2)
    )


def pack_exchange_delete_parameters(
        exchange_name: ExchangeName, if_unused: bool, no_wait: bool, encoding: str) -> bytes:
    bio = BytesIO()
    bio.write(RESERVED_SHORT_UINT_DATA)
    build_shortstr(bio, exchange_name, encoding)
    bio.write(pack_booleans(if_unused, no_wait))
    return bio.getvalue()


def parse_exchange_bind_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> ExchangeBindParameters:
    parse_binary_shortstr(bio)  # reserved
    return ExchangeBindParameters(
        parse_shortstr(bio, encoding),
        parse_shortstr(bio, encoding),
        parse_shortstr(bio, encoding),
        parse_boolean(bio),
        parse_field_table(bio, encoding, rabbitmq)
    )


def pack_exchange_bind_parameters(
        destination: ExchangeName, source: ExchangeName, routing_key: RoutingKey, no_wait: bool,
        arguments: FieldTable, encoding: str, rabbitmq: bool) -> bytes:
    bio = BytesIO()
    bio.write(RESERVED_SHORT_UINT_DATA)
    build_shortstr(bio, destination, encoding)
    build_shortstr(bio, source, encoding)
    build_shortstr(bio, routing_key, encoding)
    bio.write(pack_boolean(no_wait))
    build_field_table(bio, arguments, encoding, rabbitmq)
    return bio.getvalue()


EXCHANGE_METHOD_PARAMETER_PARSERS = {
    m.METHOD_EXCHANGE_DECLARE: parse_exchange_declare_parameters,
    m.METHOD_EXCHANGE_DECLARE_OK: None,
    m.METHOD_EXCHANGE_DELETE: parse_exchange_delete_parameters,
    m.METHOD_EXCHANGE_DELETE_OK: None,
    m.METHOD_EXCHANGE_BIND: parse_exchange_bind_parameters,
    m.METHOD_EXCHANGE_BIND_OK: None,
}



class ConsumerCancelled(Exception):
    def __init__(self):
        raise Exception('Do not instantiate me directly')


class ServerCancelledConsumer(ConsumerCancelled):
    __slots__ = ()

    def __init__(self):
        pass

    def __str__(self):
        return 'Server cancelled consumer'


class ClientCancelledConsumer(ConsumerCancelled):
    pass


class ClientCancelledConsumerOK(ClientCancelledConsumer):
    __slots__ = ()

    def __init__(self):
        pass

    def __str__(self):
        return 'Client cancelled consumer without errors'


class ClientCancelledConsumerError(ClientCancelledConsumer):
    __slots__ = 'exc',

    def __init__(self, exc: Exception):
        self.exc = exc

    def __str__(self):
        return 'Client cancelled consumer because of an unhandled exception: {!r}'.format(self.exc)



from io import BytesIO
from typing import List

from .exceptions import IncompleteFrame, ShortstrDecodingError, LongstrDecodingError
from .structs import parse_octet_uint, parse_long_uint, pack_octet_uint, pack_long_uint


def parse_shortstr(bio: BytesIO, encoding: str) -> str:
    size = parse_octet_uint(bio)
    data = bio.read(size)
    if len(data) != size:
        raise IncompleteFrame()
    try:
        return data.decode(encoding)
    except UnicodeDecodeError as exc:
        raise ShortstrDecodingError(data, encoding) from exc


def parse_split_shortstr(bio: BytesIO, encoding: str) -> List[str]:
    return parse_shortstr(bio, encoding).split(' ')


def pack_shortstr(text: str, encoding: str) -> bytes:
    data = text.encode(encoding)
    return pack_octet_uint(len(data)) + data


def build_shortstr(bio: BytesIO, text: str, encoding: str) -> None:
    data = text.encode(encoding)
    bio.write(pack_octet_uint(len(data)))
    bio.write(data)


def build_split_shortstr(bio: BytesIO, split_text: List[str], encoding: str) -> None:
    build_shortstr(bio, ' '.join(split_text), encoding)


EMPTY_SHORTSTR_DATA = pack_octet_uint(0)  # empty string constant


def parse_binary_shortstr(bio: BytesIO) -> bytes:
    size = parse_octet_uint(bio)
    data = bio.read(size)
    if len(data) != size:
        raise IncompleteFrame()
    return data


def build_binary_shortstr(bio: BytesIO, data: bytes) -> None:
    bio.write(pack_octet_uint(len(data)))
    bio.write(data)


def parse_longstr(bio: BytesIO, encoding: str) -> str:
    size = parse_long_uint(bio)
    data = bio.read(size)
    if len(data) != size:
        raise IncompleteFrame()
    try:
        return data.decode(encoding)
    except UnicodeDecodeError as exc:
        raise LongstrDecodingError(data, encoding) from exc


def parse_split_longstr(bio: BytesIO, encoding: str) -> List[str]:
    return parse_longstr(bio, encoding).split(' ')


def parse_binary_longstr(bio: BytesIO) -> bytes:
    size = parse_long_uint(bio)
    data = bio.read(size)
    if len(data) != size:
        raise IncompleteFrame()
    return data


def pack_binary_longstr(data: bytes) -> bytes:
    return pack_long_uint(len(data)) + data


def build_binary_longstr(bio: BytesIO, data: bytes) -> None:
    bio.write(pack_long_uint(len(data)))
    bio.write(data)


def build_longstr(bio: BytesIO, text: str, encoding: str) -> None:
    data = text.encode(encoding)
    bio.write(pack_long_uint(len(data)))
    bio.write(data)



from ammoo.wire.low.containers import FieldTable
from ammoo.validatation import validate_int, validate_routing_key


def add_int_arg(name: str, value: int, argument_name: str, arguments: FieldTable, zero_allowed: bool) -> None:
    """Used for adding the x-message-ttl and x-expires arguments to queue.declare"""
    validate_int(value, name)
    if zero_allowed:
        if value < 0:
            raise ValueError('{} is negative: {}'.format(name, value))
    elif value <= 0:
        raise ValueError('{} is non-positive: {}'.format(name, value))
    arguments[argument_name] = value


def add_routing_keys_header(items, arg_name, headers, header_name):
    """Utility function for publish()"""
    if isinstance(items, str):
        items = [items]
    elif not isinstance(items, list):
        raise TypeError('{} is not str or list: {!r}'.format(arg_name, items))
    for item in items:
        validate_routing_key(item, '{} item'.format(arg_name))
    headers[header_name] = items



from io import BytesIO
from struct import unpack, error as struct_error, pack

from .exceptions import IncompleteFrame, InvalidFloat, InvalidDouble


def parse_octet_uint(bio: BytesIO) -> int:
    try:
        return unpack('!B', bio.read(1))[0]
    except struct_error as exc:
        raise IncompleteFrame() from exc


def pack_octet_uint(value: int) -> bytes:
    return pack('!B', value)


def build_octet_uint(bio: BytesIO, value: int) -> None:
    bio.write(pack('!B', value))


def parse_octet_int(bio: BytesIO) -> int:
    try:
        return unpack('!b', bio.read(1))[0]
    except struct_error as exc:
        raise IncompleteFrame() from exc


def pack_octet_int(value: int) -> bytes:
    return pack('!b', value)


def parse_short_uint(bio: BytesIO) -> int:
    try:
        return unpack('!H', bio.read(2))[0]
    except struct_error as exc:
        raise IncompleteFrame() from exc


def pack_short_uint(value: int) -> bytes:
    return pack('!H', value)


RESERVED_SHORT_UINT_DATA = pack_short_uint(0)
ZERO_SHORT_UINT_DATA = pack_short_uint(0)


def parse_short_int(bio: BytesIO) -> int:
    try:
        return unpack('!h', bio.read(2))[0]
    except struct_error as exc:
        raise IncompleteFrame() from exc


def pack_short_int(value: int) -> bytes:
    return pack('!h', value)


def parse_long_uint(bio: BytesIO) -> int:
    try:
        return unpack('!I', bio.read(4))[0]
    except struct_error as exc:
        raise IncompleteFrame() from exc


def pack_long_uint(value: int) -> bytes:
    return pack('!I', value)


def parse_long_int(bio: BytesIO) -> int:
    try:
        return unpack('!i', bio.read(4))[0]
    except struct_error as exc:
        raise IncompleteFrame() from exc



def pack_long_int(value: int) -> bytes:
    return pack('!i', value)


def parse_long_long_uint(bio: BytesIO) -> int:
    try:
        return unpack('!Q', bio.read(8))[0]
    except struct_error as exc:
        raise IncompleteFrame() from exc


def pack_long_long_uint(value: int) -> bytes:
    return pack('!Q', value)


def parse_long_long_int(bio: BytesIO) -> int:
    try:
        return unpack('!q', bio.read(8))[0]
    except struct_error as exc:
        raise IncompleteFrame() from exc


def pack_long_long_int(value: int) -> bytes:
    return pack('!q', value)


def parse_float(bio: BytesIO) -> float:
    data = bio.read(4)
    try:
        return unpack('!f', data)[0]
    except struct_error as exc:
        if len(data) != 4:
            raise IncompleteFrame() from exc
        else:
            raise InvalidFloat(data) from exc


def parse_double(bio: BytesIO) -> float:
    data = bio.read(8)
    try:
        return unpack('!d', data)[0]
    except struct_error as exc:
        if len(data) != 8:
            raise IncompleteFrame() from exc
        else:
            raise InvalidDouble(data) from exc


def pack_double(value: float) -> bytes:
    return pack('!d', value)



from io import BytesIO

from ammoo.auth.abc import AuthenticationMechanism
from ammoo.wire.low.containers import build_lengthless_field_table


class AMQPlainAuth(AuthenticationMechanism):
    @classmethod
    def name(self):
        return 'AMQPLAIN'

    def __init__(self, login, password):
        if not isinstance(login, (bytes, str)):
            raise TypeError('login must be a str or bytes object')
        if not isinstance(password, (bytes, str)):
            raise TypeError('password must be a str or bytes object')
        self.table = {
            'LOGIN': login,
            'PASSWORD': password,
        }

    def build_start_ok_response(self, encoding, rabbitmq):
        bio = BytesIO()
        build_lengthless_field_table(bio, self.table, encoding, rabbitmq)
        return bio.getvalue()



from io import BytesIO
from typing import NamedTuple

from ammoo.wire import methods as m
from ammoo.wire.low.containers import parse_field_table, FieldTable, build_field_table
from ammoo.wire.low.misc import parse_boolean, parse_booleans, pack_boolean, pack_booleans
from ammoo.wire.low.strings import parse_shortstr, parse_binary_shortstr, build_shortstr, EMPTY_SHORTSTR_DATA
from ammoo.wire.low.structs import parse_short_uint, parse_long_uint, parse_long_long_uint, pack_long_long_uint, \
    pack_long_uint, pack_short_uint, RESERVED_SHORT_UINT_DATA
from ammoo.wire.typing import QueueName, ConsumerTag, ExchangeName, RoutingKey, ReplyCode, DeliveryTag, MessageCount, \
    Requeue

BasicQosParameters = NamedTuple('BasicQosParameters', [
    ('prefetch_size', int),
    ('prefetch_count', int),
    ('global_', bool),
])
BasicConsumeParameters = NamedTuple('BasicConsumeParameters', [
    ('queue_name', QueueName),
    ('consumer_tag', ConsumerTag),
    ('no_local', bool),
    ('no_ack', bool),
    ('exclusive', bool),
    ('no_wait', bool),
    ('arguments', FieldTable),
])
BasicCancelParameters = NamedTuple('BasicCancelParameters', [
    ('consumer_tag', ConsumerTag),
    ('no_wait', bool),
])
BasicPublishParameters = NamedTuple('BasicPublishParameters', [
    ('exchange_name', ExchangeName),
    ('routing_key', RoutingKey),
    ('mandatory', bool),
    ('immediate', bool)
])
BasicReturnParameters = NamedTuple('BasicReturnParameters', [
    ('reply_code', ReplyCode),
    ('reply_text', str),
    ('exchange_name', ExchangeName),
    ('routing_key', RoutingKey)
])
BasicDeliverParameters = NamedTuple('BasicDeliverParameters', [
    ('consumer_tag', ConsumerTag),
    ('delivery_tag', DeliveryTag),
    ('redelivered', bool),
    ('exchange_name', ExchangeName),
    ('routing_key', RoutingKey)
])
BasicGetParameters = NamedTuple('BasicGetParameters', [
    ('queue_name', QueueName),
    ('no_ack', bool)
])
BasicGetOkParameters = NamedTuple('BasicGetOkParameters', [
    ('delivery_tag', DeliveryTag),
    ('redelivered', bool),
    ('exchange_name', ExchangeName),
    ('routing_key', RoutingKey),
    ('message_count', MessageCount)
])
BasicAckParameters = NamedTuple('BasicAckParameters', [
    ('delivery_tag', DeliveryTag),
    ('multiple', bool),
])
BasicRejectParameters = NamedTuple('BasicRejectParameters', [
    ('delivery_tag', DeliveryTag),
    ('requeue', bool),
])
BasicNackParameters = NamedTuple('BasicNackParameters', [
    ('delivery_tag', DeliveryTag),
    ('multiple', bool),
    ('requeue', bool),
])


def parse_basic_qos_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> BasicQosParameters:
    return BasicQosParameters(
        parse_long_uint(bio),
        parse_short_uint(bio),
        parse_boolean(bio)
    )


def pack_basic_qos_parameters(prefetch_size: int, prefetch_count: int, global_: bool) -> bytes:
    return pack_long_uint(prefetch_size) + pack_short_uint(prefetch_count) + pack_boolean(global_)


def parse_basic_consume_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> BasicConsumeParameters:
    parse_binary_shortstr(bio)
    return BasicConsumeParameters(
        parse_shortstr(bio, encoding),
        parse_shortstr(bio, encoding),
        *parse_booleans(bio, 4),
        parse_field_table(bio, encoding, rabbitmq)
    )


def pack_basic_consume_parameters(
        queue_name: QueueName, consumer_tag: ConsumerTag, no_local: bool, no_ack: bool, exclusive: bool, no_wait: bool,
        arguments: FieldTable, encoding: str, rabbitmq: bool) -> bytes:
    bio = BytesIO()
    bio.write(RESERVED_SHORT_UINT_DATA)
    build_shortstr(bio, queue_name, encoding)
    build_shortstr(bio, consumer_tag, encoding)
    bio.write(pack_booleans(no_local, no_ack, exclusive, no_wait))
    build_field_table(bio, arguments, encoding, rabbitmq)
    return bio.getvalue()


def parse_basic_consume_ok_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> ConsumerTag:
    return parse_shortstr(bio, encoding)


def parse_basic_cancel_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> BasicCancelParameters:
    return BasicCancelParameters(
        parse_shortstr(bio, encoding),
        parse_boolean(bio)
    )


def pack_basic_cancel_parameters(consumer_tag: ConsumerTag, no_wait: bool, encoding: str) -> bytes:
    bio = BytesIO()
    build_shortstr(bio, consumer_tag, encoding)
    bio.write(pack_boolean(no_wait))
    return bio.getvalue()


def parse_basic_cancel_ok_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> ConsumerTag:
    return parse_shortstr(bio, encoding)


def parse_basic_publish_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> BasicPublishParameters:
    parse_binary_shortstr(bio)
    return BasicPublishParameters(
        parse_shortstr(bio, encoding),
        parse_shortstr(bio, encoding),
        *parse_booleans(bio, 2)
    )


def pack_basic_publish_parameters(
        exchange_name: ExchangeName, routing_key: RoutingKey, mandatory: bool, immediate: bool, encoding: str) -> bytes:
    bio = BytesIO()
    bio.write(RESERVED_SHORT_UINT_DATA)
    build_shortstr(bio, exchange_name, encoding)
    build_shortstr(bio, routing_key, encoding)
    bio.write(pack_booleans(mandatory, immediate))
    return bio.getvalue()


def parse_basic_return_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> BasicReturnParameters:
    return BasicReturnParameters(
        parse_short_uint(bio),
        parse_shortstr(bio, encoding),
        parse_shortstr(bio, encoding),
        parse_shortstr(bio, encoding)
    )


def parse_basic_deliver_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> BasicDeliverParameters:
    return BasicDeliverParameters(
        parse_shortstr(bio, encoding),
        parse_long_long_uint(bio),
        parse_boolean(bio),
        parse_shortstr(bio, encoding),
        parse_shortstr(bio, encoding)
    )


def parse_basic_get_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> BasicGetParameters:
    parse_binary_shortstr(bio)
    return BasicGetParameters(
        parse_shortstr(bio, encoding),
        parse_boolean(bio)
    )


def pack_basic_get_parameters(queue_name: QueueName, no_ack: bool, encoding: str) -> bytes:
    bio = BytesIO()
    bio.write(RESERVED_SHORT_UINT_DATA)
    build_shortstr(bio, queue_name, encoding)
    bio.write(pack_boolean(no_ack))
    return bio.getvalue()


def parse_basic_get_ok_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> BasicGetOkParameters:
    return BasicGetOkParameters(
        parse_long_long_uint(bio),
        parse_boolean(bio),
        parse_shortstr(bio, encoding),
        parse_shortstr(bio, encoding),
        parse_long_uint(bio)
    )


def parse_basic_get_empty_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> None:
    parse_binary_shortstr(bio)
    return None


def parse_basic_ack_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> BasicAckParameters:
    return BasicAckParameters(
        parse_long_long_uint(bio),
        parse_boolean(bio)
    )


def pack_basic_ack_parameters(delivery_tag: DeliveryTag, multiple: bool) -> bytes:
    return pack_long_long_uint(delivery_tag) + pack_boolean(multiple)


def parse_basic_reject_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> BasicRejectParameters:
    return BasicRejectParameters(
        parse_long_long_uint(bio),
        parse_boolean(bio)
    )


def pack_basic_reject_parameters(delivery_tag: DeliveryTag, requeue: bool) -> bytes:
    return pack_long_long_uint(delivery_tag) + pack_boolean(requeue)


# used for basic.recover-async too
def parse_basic_recover_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> Requeue:
    return parse_boolean(bio)


def parse_basic_nack_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> BasicNackParameters:
    return BasicNackParameters(
        parse_long_long_uint(bio),
        *parse_booleans(bio, 2)
    )


def pack_basic_nack_parameters(delivery_tag: DeliveryTag, multiple: bool, requeue: bool) -> bytes:
    return pack_long_long_uint(delivery_tag) + pack_booleans(multiple, requeue)


BASIC_METHOD_PARAMETER_PARSERS = {
    m.METHOD_BASIC_QOS: parse_basic_qos_parameters,
    m.METHOD_BASIC_QOS_OK: None,
    m.METHOD_BASIC_CONSUME: parse_basic_consume_parameters,
    m.METHOD_BASIC_CONSUME_OK: parse_basic_consume_ok_parameters,
    m.METHOD_BASIC_CANCEL: parse_basic_cancel_parameters,
    m.METHOD_BASIC_CANCEL_OK: parse_basic_cancel_ok_parameters,
    m.METHOD_BASIC_PUBLISH: parse_basic_publish_parameters,
    m.METHOD_BASIC_RETURN_: parse_basic_return_parameters,
    m.METHOD_BASIC_DELIVER: parse_basic_deliver_parameters,
    m.METHOD_BASIC_GET: parse_basic_get_parameters,
    m.METHOD_BASIC_GET_OK: parse_basic_get_ok_parameters,
    m.METHOD_BASIC_GET_EMPTY: parse_basic_get_empty_parameters,
    m.METHOD_BASIC_ACK: parse_basic_ack_parameters,
    m.METHOD_BASIC_REJECT: parse_basic_reject_parameters,
    m.METHOD_BASIC_RECOVER_ASYNC: parse_basic_recover_parameters,
    m.METHOD_BASIC_RECOVER: parse_basic_recover_parameters,
    m.METHOD_BASIC_RECOVER_OK: None,
    m.METHOD_BASIC_NACK: parse_basic_nack_parameters,
}



from datetime import datetime
from io import BytesIO
from typing import Dict, Union, List
from collections.abc import Mapping as MappingABC

from .exceptions import IncompleteFrame, InvalidFieldValueType, InvalidBooleanTable
from .misc import parse_boolean, parse_decimal, parse_timestamp, parse_no_field, pack_boolean, pack_timestamp
from .strings import parse_shortstr, parse_longstr, build_shortstr, build_longstr, build_binary_longstr, \
    parse_binary_longstr
from .structs import parse_long_uint, parse_octet_uint, parse_short_uint, parse_long_long_uint, parse_octet_int, \
    parse_short_int, parse_long_int, parse_long_long_int, parse_float, parse_double, pack_long_uint, pack_octet_uint, \
    pack_short_uint, pack_long_long_uint, pack_long_long_int, pack_octet_int, pack_short_int, pack_long_int, pack_double


FieldValue = Union[int, float, str, datetime, bool, 'FieldTable', 'FieldTable', None]
FieldTable = Dict[str, FieldValue]
FieldArray = List[FieldValue]


def parse_field_value(bio: BytesIO, encoding: str, rabbitmq: bool) -> FieldValue:
    value_type = bio.read(1)
    if rabbitmq:
        field_value_types = RABBITMQ_FIELD_VALUE_TYPES
    else:
        field_value_types = FIELD_VALUE_TYPES
    try:
        parse_value, args_type = field_value_types[value_type]
    except KeyError:
        if len(value_type) == 0:
            raise IncompleteFrame()
        raise InvalidFieldValueType(value_type)
    if args_type == NO_ARGS:
        return parse_value(bio)
    elif args_type == STRING_ARGS:
        return parse_value(bio, encoding)
    else:
        return parse_value(bio, encoding, rabbitmq)


def parse_field_anystring(bio: BytesIO, encoding: str, rabbitmq: bool) -> str:
    value_type = bio.read(1)
    if rabbitmq:
        field_value_types = RABBITMQ_STRING_FIELD_VALUE_TYPES
    else:
        field_value_types = STRING_FIELD_VALUE_TYPES
    try:
        parse_value = field_value_types[value_type]
    except KeyError:
        if len(value_type) == 0:
            raise IncompleteFrame()
        raise InvalidFieldValueType(value_type)
    return parse_value(bio, encoding)


def parse_field_table(bio: BytesIO, encoding: str, rabbitmq: bool) -> FieldTable:
    length = parse_long_uint(bio)
    table = {}  # could be OrderedDict, but we don't really care about the order
    end_offset = bio.tell() + length
    while bio.tell() < end_offset:
        name = parse_shortstr(bio, encoding)
        value = parse_field_value(bio, encoding, rabbitmq)
        table[name] = value
    return table


def build_field_table(bio: BytesIO, table: FieldTable, encoding: str, rabbitmq: bool) -> None:
    table_bio = BytesIO()
    for name, value in table.items():
        build_shortstr(table_bio, name, encoding)
        pack_field_value(table_bio, value, encoding, rabbitmq)
    bio.write(pack_long_uint(table_bio.tell()))
    bio.write(table_bio.getvalue())


def build_lengthless_field_table(bio: BytesIO, table: FieldTable, encoding: str, rabbitmq: bool) -> None:
    for name, value in table.items():
        build_shortstr(bio, name, encoding)
        pack_field_value(bio, value, encoding, rabbitmq)


def parse_boolean_table_field(bio: BytesIO, encoding: str) -> Dict[str, bool]:
    value_type = bio.read(1)
    if value_type != b'F':
        if len(value_type) == 0:
            raise IncompleteFrame()
        raise InvalidFieldValueType(value_type)
    length = parse_long_uint(bio)
    table = {}  # could be OrderedDict, but we don't really care about the order
    end_offset = bio.tell() + length
    while bio.tell() < end_offset:
        name = parse_shortstr(bio, encoding)
        field_value_type = bio.read(1)
        if field_value_type != b't':
            if field_value_type == b'':
                raise IncompleteFrame()
            else:
                raise InvalidBooleanTable(field_value_type)
        table[name] = parse_boolean(bio)
    return table


def build_boolean_table_field(bio: BytesIO, table: Dict[str, bool], encoding: str) -> None:
    bio.write(b'F')
    table_bio = BytesIO()
    for name, value in table.items():
        build_shortstr(table_bio, name, encoding)
        table_bio.write(b't')
        table_bio.write(pack_boolean(value))
    bio.write(pack_long_uint(table_bio.tell()))
    bio.write(table_bio.getvalue())


def parse_field_array(bio: BytesIO, encoding: str, rabbitmq: bool) -> FieldArray:
    length = parse_long_uint(bio)
    array = []
    end_offset = bio.tell() + length
    while bio.tell() < end_offset:
        value = parse_field_value(bio, encoding, rabbitmq)
        array.append(value)
    return array


def build_field_array(bio: BytesIO, items: FieldArray, encoding: str, rabbitmq: bool) -> None:
    array_bio = BytesIO()
    for item in items:
        pack_field_value(array_bio, item, encoding, rabbitmq)
    bio.write(pack_long_uint(array_bio.tell()))
    bio.write(array_bio.getvalue())


NO_ARGS = 1
FIELD_ARGS = 2
STRING_ARGS = 3

STRING_FIELD_VALUE_TYPES = {
    b's': parse_shortstr,  # rabbitmq: N/A
    b'S': parse_longstr,
}


FIELD_VALUE_TYPES = {
    b't': (parse_boolean, NO_ARGS),
    b'b': (parse_octet_int, NO_ARGS),
    b'B': (parse_octet_uint, NO_ARGS),
    b'U': (parse_short_int, NO_ARGS),  # rabbitmq: (s
    b'u': (parse_short_uint, NO_ARGS),
    b'I': (parse_long_int, NO_ARGS),
    b'i': (parse_long_uint, NO_ARGS),
    b'L': (parse_long_long_int, NO_ARGS),  # rabbitmq: (l
    b'l': (parse_long_long_uint, NO_ARGS),  # rabbitmq: (N/A
    b'f': (parse_float, NO_ARGS),
    b'd': (parse_double, NO_ARGS),
    b'D': (parse_decimal, NO_ARGS),
    b'T': (parse_timestamp, NO_ARGS),
    b'V': (parse_no_field, NO_ARGS),
    b's': (parse_shortstr, STRING_ARGS),  # rabbitmq: N/A
    b'S': (parse_longstr, STRING_ARGS),
    b'A': (parse_field_array, FIELD_ARGS),
    b'F': (parse_field_table, FIELD_ARGS),
}


RABBITMQ_STRING_FIELD_VALUE_TYPES = {
    b'S': parse_longstr,
}


#  https://www.rabbitmq.com/amqp-0-9-1-errata.html / 3 Field types
RABBITMQ_FIELD_VALUE_TYPES = {
    b't': (parse_boolean, NO_ARGS),
    b'b': (parse_octet_int, NO_ARGS),
    b'B': (parse_octet_uint, NO_ARGS),
    b's': (parse_short_int, NO_ARGS),  # differs
    b'u': (parse_short_uint, NO_ARGS),
    b'I': (parse_long_int, NO_ARGS),
    b'i': (parse_long_uint, NO_ARGS),
    b'l': (parse_long_long_int, NO_ARGS),  # differs
    # no Unsigned 64-bit
    b'f': (parse_float, NO_ARGS),
    b'd': (parse_double, NO_ARGS),
    b'D': (parse_decimal, NO_ARGS),
    b'T': (parse_timestamp, NO_ARGS),
    b'V': (parse_no_field, NO_ARGS),
    # no short string
    b'S': (parse_longstr, STRING_ARGS),
    b'A': (parse_field_array, FIELD_ARGS),
    b'F': (parse_field_table, FIELD_ARGS),
    b'x': (parse_binary_longstr, NO_ARGS)  # rabbitmq extension
}


MIN_8BIT = -2 ** 7
MIN_16BIT = -2 ** 15
MIN_32BIT = -2 ** 31
TWO_POW_16 = 2 ** 16
TWO_POW_32 = 2 ** 32


#FieldValue = Union[int, float, str, datetime, bool, 'FieldTable', 'FieldTable']
def pack_field_value(bio: BytesIO, value: FieldValue, encoding: str, rabbitmq: bool):
    if isinstance(value, str):
        # on non-rabbitmq we could use shortstr if binary length < 256 bytes
        bio.write(b'S')
        build_longstr(bio, value, encoding)
    elif isinstance(value, bytes):
        # on non-rabbitmq we could use shortstr if binary length < 256 bytes
        # and on rabbitmq use type byte 'x'
        bio.write(b'S')
        build_binary_longstr(bio, value)
    elif isinstance(value, bool):
        bio.write(b't')
        bio.write(pack_boolean(value))
    elif isinstance(value, int):
        if value >= 0:
            if value < 256:
                bio.write(b'B')
                bio.write(pack_octet_uint(value))
            elif value < TWO_POW_16:
                bio.write(b'u')
                bio.write(pack_short_uint(value))
            elif value < TWO_POW_32:
                bio.write(b'i')
                bio.write(pack_long_uint(value))
            else:  # 64-bit
                bio.write(b'l')
                if rabbitmq:
                    bio.write(pack_long_long_int(value))
                else:
                    bio.write(pack_long_long_uint(value))
        else:
            if value >= MIN_8BIT:
                bio.write(b'b')
                bio.write(pack_octet_int(value))
            elif value >= MIN_16BIT:
                if rabbitmq:
                    bio.write(b's')
                else:
                    bio.write(b'U')
                bio.write(pack_short_int(value))
            elif value >= MIN_32BIT:
                bio.write(b'I')
                bio.write(pack_long_int(value))
            else:  # 64-bit
                if rabbitmq:
                    bio.write(b'l')
                else:
                    bio.write(b'L')
                bio.write(pack_long_long_int(value))
    elif isinstance(value, float):
        bio.write(b'd')
        bio.write(pack_double(value))
    elif isinstance(value, MappingABC):
        bio.write(b'F')
        build_field_table(bio, value, encoding, rabbitmq)
    elif isinstance(value, list):
        bio.write(b'A')
        build_field_array(bio, value, encoding, rabbitmq)
    elif isinstance(value, datetime):
        bio.write(b'T')
        bio.write(pack_timestamp(value))
    elif value is None:
        bio.write(b'V')
    else:
        raise TypeError(value)



from .version import __version__
from .connect import connect
from .connection import Connection
from .channel import Channel
from .consumer import Consumer, ConsumerIterator
from .message import Message, ExpectedMessage, DeliverMessage, GetMessage, ReturnMessage
from .exceptions.connection import ConnectionClosed
from .exceptions.channel import ChannelClosed, EmptyQueue



from io import BytesIO
from typing import NamedTuple

from ammoo.wire import methods as m
from ammoo.wire.low.containers import parse_field_table, FieldTable, build_field_table
from ammoo.wire.low.misc import parse_booleans, parse_boolean, pack_booleans, pack_boolean
from ammoo.wire.low.strings import parse_binary_shortstr, parse_shortstr, build_shortstr
from ammoo.wire.low.structs import parse_long_uint, RESERVED_SHORT_UINT_DATA
from ammoo.wire.typing import MessageCount, QueueName, ConsumerCount, ExchangeName, RoutingKey


QueueDeclareParameters = NamedTuple('QueueDeclareParameters', [
    ('queue_name', QueueName),
    ('passive', bool),
    ('durable', bool),
    ('exclusive', bool),
    ('auto_delete', bool),
    ('no_wait', bool),
    ('arguments', FieldTable),
])
QueueDeclareOkParameters = NamedTuple('QueueDeclareOkParameters', [
    ('queue_name', QueueName),
    ('message_count', MessageCount),
    ('consumer_count', ConsumerCount),
])
QueueBindParameters = NamedTuple('QueueBindParameters', [
    ('queue_name', QueueName),
    ('exchange_name', ExchangeName),
    ('routing_key', RoutingKey),
    ('no_wait', bool),
    ('arguments', FieldTable),
])
QueuePurgeParameters = NamedTuple('QueuePurgeParameters', [
    ('queue_name', QueueName),
    ('no_wait', bool),
])
QueueDeleteParameters = NamedTuple('QueueDeleteParameters', [
    ('queue_name', QueueName),
    ('if_unused', bool),
    ('if_empty', bool),
    ('no_wait', bool),
])
QueueUnbindParameters = NamedTuple('QueueUnbindParameters', [
    ('queue_name', QueueName),
    ('exchange_name', ExchangeName),
    ('routing_key', RoutingKey),
    ('arguments', FieldTable),
])


def parse_queue_declare_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> QueueDeclareParameters:
    parse_binary_shortstr(bio)
    return QueueDeclareParameters(
        parse_shortstr(bio, encoding),
        parse_booleans(bio, 5),
        parse_field_table(bio, encoding, rabbitmq)
    )


def pack_queue_declare_parameters(
        queue_name: QueueName, passive: bool, durable: bool, exclusive: bool, auto_delete: bool, no_wait: bool,
        arguments: FieldTable, encoding: str, rabbitmq: bool) -> bytes:
    bio = BytesIO()
    bio.write(RESERVED_SHORT_UINT_DATA)
    build_shortstr(bio, queue_name, encoding)
    bio.write(pack_booleans(passive, durable, exclusive, auto_delete, no_wait))
    build_field_table(bio, arguments, encoding, rabbitmq)
    return bio.getvalue()


def parse_queue_declare_ok_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> QueueDeclareOkParameters:
    return QueueDeclareOkParameters(
        parse_shortstr(bio, encoding),
        parse_long_uint(bio),
        parse_long_uint(bio)
    )


def parse_queue_bind_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> QueueBindParameters:
    parse_binary_shortstr(bio)
    return QueueBindParameters(
        parse_shortstr(bio, encoding),
        parse_shortstr(bio, encoding),
        parse_shortstr(bio, encoding),
        parse_boolean(bio),
        parse_field_table(bio, encoding, rabbitmq)
    )


def pack_queue_bind_parameters(
        queue_name: QueueName, exchange_name: ExchangeName, routing_key: RoutingKey, no_wait: bool,
        arguments: FieldTable, encoding: str, rabbitmq: bool) -> bytes:
    bio = BytesIO()
    bio.write(RESERVED_SHORT_UINT_DATA)
    build_shortstr(bio, queue_name, encoding)
    build_shortstr(bio, exchange_name, encoding)
    build_shortstr(bio, routing_key, encoding)
    bio.write(pack_boolean(no_wait))
    build_field_table(bio, arguments, encoding, rabbitmq)
    return bio.getvalue()


def parse_queue_purge_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> QueuePurgeParameters:
    parse_binary_shortstr(bio)
    return QueuePurgeParameters(
        parse_shortstr(bio, encoding),
        parse_boolean(bio)
    )


def pack_queue_purge_parameters(queue_name: QueueName, no_wait: bool, encoding: str) -> bytes:
    bio = BytesIO()
    bio.write(RESERVED_SHORT_UINT_DATA)
    build_shortstr(bio, queue_name, encoding)
    bio.write(pack_boolean(no_wait))
    return bio.getvalue()


def parse_queue_purge_ok_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> MessageCount:
    return parse_long_uint(bio)


def parse_queue_delete_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> QueueDeleteParameters:
    parse_binary_shortstr(bio)
    return QueueDeleteParameters(
        parse_shortstr(bio, encoding),
        *parse_booleans(bio, 2),
        parse_field_table(bio, encoding, rabbitmq)
    )


def pack_queue_delete_parameters(
        queue_name: QueueName, if_unused: bool, if_empty: bool, no_wait: bool, encoding: str) -> bytes:
    bio = BytesIO()
    bio.write(RESERVED_SHORT_UINT_DATA)
    build_shortstr(bio, queue_name, encoding)
    bio.write(pack_booleans(if_unused, if_empty, no_wait))
    return bio.getvalue()


def parse_queue_delete_ok_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> MessageCount:
    return parse_long_uint(bio)


def parse_queue_unbind_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> QueueUnbindParameters:
    parse_binary_shortstr(bio)
    return QueueUnbindParameters(
        parse_shortstr(bio, encoding),
        parse_shortstr(bio, encoding),
        parse_shortstr(bio, encoding),
        parse_field_table(bio, encoding, rabbitmq)
    )


def pack_queue_unbind_parameters(
        queue_name: QueueName, exchange_name: ExchangeName, routing_key: RoutingKey, arguments: FieldTable,
        encoding: str, rabbitmq: bool) -> bytes:
    bio = BytesIO()
    bio.write(RESERVED_SHORT_UINT_DATA)
    build_shortstr(bio, queue_name, encoding)
    build_shortstr(bio, exchange_name, encoding)
    build_shortstr(bio, routing_key, encoding)
    build_field_table(bio, arguments, encoding, rabbitmq)
    return bio.getvalue()


QUEUE_METHOD_PARAMETER_PARSERS = {
    m.METHOD_QUEUE_DECLARE: parse_queue_declare_parameters,
    m.METHOD_QUEUE_DECLARE_OK: parse_queue_declare_ok_parameters,
    m.METHOD_QUEUE_BIND: parse_queue_bind_parameters,
    m.METHOD_QUEUE_BIND_OK: None,
    m.METHOD_QUEUE_PURGE: parse_queue_purge_parameters,
    m.METHOD_QUEUE_PURGE_OK: parse_queue_purge_ok_parameters,
    m.METHOD_QUEUE_DELETE: parse_queue_delete_parameters,
    m.METHOD_QUEUE_DELETE_OK: parse_queue_delete_ok_parameters,
    m.METHOD_QUEUE_UNBIND: parse_queue_unbind_parameters,
    m.METHOD_QUEUE_UNBIND_OK: None,
}



from io import BytesIO
from struct import pack
from typing import NamedTuple

from ammoo.wire import methods as m
from ammoo.wire.low.misc import parse_boolean
from ammoo.wire.low.strings import parse_shortstr, parse_binary_shortstr, parse_binary_longstr
from ammoo.wire.low.structs import parse_short_uint
from ammoo.wire.typing import ReplyCode, FlowActive


ChannelCloseParameters = NamedTuple('ChannelCloseParameters', [
    ('reply_code', ReplyCode),
    ('reply_text', str),
    ('class_id', int),
    ('method_id', int),
])


CLEAN_CHANNEL_CLOSE_PARAMETERS_DATA = pack('!HBHH', 0, 0, 0, 0)


def parse_channel_open_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> None:
    parse_binary_shortstr(bio)
    return None


def parse_channel_open_ok_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> None:
    parse_binary_longstr(bio)
    return None


# used for channel.flow-ok too
def parse_channel_flow_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> FlowActive:
    return parse_boolean(bio)


def parse_channel_close_parameters(bio: BytesIO, encoding: str, rabbitmq: bool) -> ChannelCloseParameters:
    return ChannelCloseParameters(
        parse_short_uint(bio),
        parse_shortstr(bio, encoding),
        parse_short_uint(bio),
        parse_short_uint(bio),
    )


CHANNEL_METHOD_PARAMETER_PARSERS = {
    m.METHOD_CHANNEL_OPEN: parse_channel_open_parameters,
    m.METHOD_CHANNEL_OPEN_OK: parse_channel_open_ok_parameters,
    m.METHOD_CHANNEL_FLOW: parse_channel_flow_parameters,
    m.METHOD_CHANNEL_FLOW_OK: parse_channel_flow_parameters,
    m.METHOD_CHANNEL_CLOSE: parse_channel_close_parameters,
    m.METHOD_CHANNEL_CLOSE_OK: None,
}



# -*- encoding: utf-8 -*-
import re
import os

from setuptools import setup

PACKAGE_NAME = 'ammoo'
PACKAGES = [
    'ammoo',
    'ammoo.auth',
    'ammoo.exceptions',
    'ammoo.wire',
    'ammoo.auth.mechanisms',
    'ammoo.wire.frames',
    'ammoo.wire.low',
    'ammoo.wire.frames.method'
]


def get_version():
    version_py_path = os.path.join(os.path.dirname(__file__), PACKAGE_NAME, 'version.py')
    pattern = re.compile(r'''^__version__\s*=\s*(?P<quote>['"])(?P<version>[^'"]+)(?P=quote)\s*(?:#.*)?$''')
    with open(version_py_path, 'r', encoding='utf-8') as fob:
        for line in fob:
            match = pattern.match(line)
            if match is not None:
                return match.group('version')
    raise ValueError('No version found')


VERSION = get_version()


setup(
    name=PACKAGE_NAME,
    version=VERSION,
    author='Mathias Fröjdman',
    author_email='mwf@iki.fi',
    url='https://github.com/mwfrojdman/ammoo',
    description='AMQP library with a pythonic API for asyncio on Python 3.5+',
    download_url='https://github.com/mwfrojdman/ammoo/archive/{}.tar.gz'.format(VERSION),
    packages=PACKAGES,
    package_data={'': ['LICENSE']},
    extras_require={
        'test': [
            'pytest',
            'pytest-asyncio',
            'pytest-timeout',
            'pytest-capturelog',
        ],
    },
    license='Apache 2.0',
    classifiers=[
        'Development Status :: 4 - Beta',
	'Framework :: AsyncIO',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    platforms='all'
)



from asyncio import Queue
from itertools import chain


class QueueClosed(Exception):
    pass


class CloseableQueue(Queue):
    def _init(self, maxsize):
        super()._init(maxsize)
        self._closed = False

    def close(self):
        self._closed = True
        for waiter in chain(self._getters, self._putters):
            if not waiter.done():
                waiter.set_exception(QueueClosed)
        self._getters.clear()
        self._putters.clear()
        self._finished.set()

    async def put(self, item):
        if self._closed:
            raise QueueClosed
        return await super().put(item)

    def put_nowait(self, item):
        if self._closed:
            raise QueueClosed
        return super().put_nowait(item)

    async def get(self):
        if self._closed:
            raise QueueClosed
        return await super().get()

    def get_nowait(self):
        if self._closed:
            raise QueueClosed
        return super().get_nowait()

    def task_done(self):
        if self._closed:
            raise QueueClosed
        return super().task_done()

    async def join(self):
        if self._closed:
            raise QueueClosed
        await super().join()
        if self._closed:
            raise QueueClosed
